#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from typing import List

from gitlabracadabra.objects.object import GitLabracadabraObject


logger = logging.getLogger(__name__)


class GitLabracadabraApplicationSettings(GitLabracadabraObject):
    EXAMPLE_YAML_HEADER = 'application_settings:\n  type: application-settings\n'
    DOC: List[str] = [
        '# Application settings lifecycle',
        'gitlab_id',
    ]
    SCHEMA = {
        '$schema': 'http://json-schema.org/draft-04/schema#',
        'title': 'Application settings',
        'type': 'object',
        'properties': {
            # Standard properties
            'gitlab_id': {
                'type': 'string',
                'description': 'GitLab id',
                '_example': 'gitlab',
                '_doc_link': 'action_file.md#gitlab_id',
            },
            # From https://docs.gitlab.com/ee/api/settings.html#list-of-settings-that-can-be-accessed-via-api-calls
            'admin_notification_email': {
                'type': 'string',
                'description': 'Abuse reports will be sent to this address if it is set. Abuse reports are always '
                               'available in the admin area.',
            },
            'after_sign_out_path': {
                'type': 'string',
                'description': 'Where to redirect users after logout.',
            },
            'after_sign_up_text': {
                'type': 'string',
                'description': 'Text shown to the user after signing up',
            },
            'akismet_api_key': {
                'type': 'string',
                'description': 'API key for akismet spam protection.',
            },
            'akismet_enabled': {
                'type': 'boolean',
                'description': 'Enable or disable akismet spam protection.',
            },
            'allow_group_owners_to_manage_ldap': {  # >= PREMIUM, SILVER
                'type': 'boolean',
                'description': 'Set to true to allow group owners to manage LDAP',
            },
            'allow_local_requests_from_hooks_and_services': {
                'type': 'boolean',
                'description': 'Allow requests to the local network from hooks and services.',
            },
            'archive_builds_in_human_readable': {
                'type': 'string',
                'description': 'Set the duration for which the jobs will be considered as old and expired. Once that '
                               'time passes, the jobs will be archived and no longer able to be retried. Make it empty '
                               'to never expire jobs. It has to be no less than 1 day, for example: `15 days`, `1 '
                               'month`, `2 years`.',
            },
            'authorized_keys_enabled': {
                'type': 'boolean',
                'description': 'By default, we write to the authorized_keys file to support Git over SSH without '
                               'additional configuration. GitLab can be optimized to authenticate SSH keys via the '
                               'database file. Only disable this if you have configured your OpenSSH server to use the '
                               'AuthorizedKeysCommand.',
            },
            'auto_devops_domain': {
                'type': 'string',
                'description': 'Specify a domain to use by default for every project’s Auto Review Apps and Auto '
                               'Deploy stages.',
            },
            'auto_devops_enabled': {
                'type': 'boolean',
                'description': 'Enable Auto DevOps for projects by default. It will automatically build, test, and '
                               'deploy applications based on a predefined CI/CD configuration.',
            },
            'check_namespace_plan': {  # >= PREMIUM, SILVER
                'type': 'boolean',
                'description': 'Enabling this will make only licensed EE features available to projects if the project '
                               'namespace’s plan includes the feature or if the project is public.',
            },
            'commit_email_hostname': {
                'type': 'string',
                'description': ' Custom hostname (for private commit emails).',
            },
            'container_registry_token_expire_delay': {
                'type': 'integer',
                'description': 'Container Registry token duration in minutes.',
            },
            'default_artifacts_expire_in': {
                'type': 'string',
                'description': 'Set the default expiration time for each job’s artifacts.',
            },
            'default_branch_protection': {
                'type': 'integer',
                'description': 'Determine if developers can push to master. Can take: 0 (not protected, both '
                               'developers and maintainers can push new commits, force push, or delete the branch), 1 '
                               '(partially protected, developers and maintainers can push new commits, but cannot '
                               'force push or delete the branch) or 2 (fully protected, developers cannot push new '
                               'commits, but maintainers can; no-one can force push or delete the branch) as a '
                               'parameter. Default is 2.',
                'enum': [0, 1, 2, 3],
            },
            'default_ci_config_path': {
                'type': 'string',
                'description': 'Default CI configuration path for new projects',
            },
            'default_group_visibility': {
                'type': 'string',
                'description': 'What visibility level new groups receive. Can take private, internal and public as a '
                               'parameter. Default is private.',
                'enum': ['private', 'internal', 'public'],
            },
            'default_project_creation': {
                'type': 'integer',
                'description': 'Default project creation protection. Can take: `0` _(No one)_, `1` _(Maintainers)_ or '
                               '`2` _(Developers + Maintainers)_',
                'enum': [0, 1, 2],
            },
            'default_project_visibility': {
                'type': 'string',
                'description': 'What visibility level new projects receive. Can take `private`, `internal` and '
                               '`public` as a parameter.',
                'enum': ['private', 'internal', 'public'],
            },
            'default_projects_limit': {
                'type': 'integer',
                'description': 'Project limit per user.',
            },
            'default_snippet_visibility': {
                'type': 'string',
                'description': 'What visibility level new snippets receive. Can take private, internal and public as a '
                               'parameter. Default is private.',
                'enum': ['private', 'internal', 'public'],
            },
            'diff_max_patch_bytes': {
                'type': 'integer',
                'description': 'Maximum diff patch size (Bytes).',
            },
            'disabled_oauth_sign_in_sources': {
                'type': 'array',
                'description': 'Disabled OAuth sign-in sources.',
                'items': {
                    'type': 'string',
                },
                'uniqueItems': True,
            },
            'dns_rebinding_protection_enabled': {
                'type': 'boolean',
                'description': 'Enforce DNS rebinding attack protection.',
            },
            'domain_blacklist': {
                'type': 'array',
                'description': 'Users with e-mail addresses that match these domain(s) will NOT be able to sign-up. '
                               'Wildcards allowed. Use separate lines for multiple entries. Ex: domain.com, '
                               '*.domain.com.',
                'items': {
                    'type': 'string',
                },
                'uniqueItems': True,
            },
            'domain_blacklist_enabled': {
                'type': 'boolean',
                'description': 'Allows blocking sign-ups from emails from specific domains.',
            },
            'domain_whitelist': {
                'type': 'array',
                'description': 'Force people to use only corporate emails for sign-up. Default is null, meaning there '
                               'is no restriction.',
                'items': {
                    'type': 'string',
                },
                'uniqueItems': True,
            },
            'dsa_key_restriction': {
                'type': 'integer',
                'description': 'The minimum allowed bit length of an uploaded DSA key. Default is 0 (no restriction). '
                               '-1 disables DSA keys.',
            },
            'ecdsa_key_restriction': {
                'type': 'integer',
                'description': 'The minimum allowed curve size (in bits) of an uploaded ECDSA key. Default is 0 (no '
                               'restriction). -1 disables ECDSA keys.',
            },
            'ed25519_key_restriction': {
                'type': 'integer',
                'description': 'The minimum allowed curve size (in bits) of an uploaded ED25519 key. Default is 0 (no '
                               'restriction). -1 disables ED25519 keys.',
            },
            'elasticsearch_aws': {  # >= PREMIUM, SILVER
                'type': 'boolean',
                'description': 'Enable the use of AWS hosted Elasticsearch',
            },
            'elasticsearch_aws_access_key': {  # >= PREMIUM, SILVER
                'type': 'string',
                'description': 'AWS IAM access key',
            },
            'elasticsearch_aws_region': {  # >= PREMIUM, SILVER
                'type': 'string',
                'description': 'The AWS region the elasticsearch domain is configured',
            },
            'elasticsearch_aws_secret_access_key': {  # >= PREMIUM, SILVER
                'type': 'string',
                'description': 'AWS IAM secret access key',
            },
            'elasticsearch_experimental_indexer': {  # >= PREMIUM, SILVER
                'type': 'boolean',
                'description': 'Use the experimental elasticsearch indexer. More info: '
                               'https://gitlab.com/gitlab-org/gitlab-elasticsearch-indexer ',
            },
            'elasticsearch_indexing': {  # >= PREMIUM, SILVER
                'type': 'boolean',
                'description': 'Enable Elasticsearch indexing',
            },
            'elasticsearch_search': {  # >= PREMIUM, SILVER
                'type': 'boolean',
                'description': 'Enable Elasticsearch search',
            },
            'elasticsearch_url': {  # >= PREMIUM, SILVER
                'type': 'string',
                'description': 'The url to use for connecting to Elasticsearch. Use a comma-separated list to support '
                               'cluster (e.g., http://localhost:9200, http://localhost:9201). If your Elasticsearch '
                               'instance is password protected, pass the username:password in the URL (e.g., '
                               'http://<username>:<password>@<elastic_host>:9200/).',
            },
            'elasticsearch_limit_indexing': {  # >= PREMIUM, SILVER
                'type': 'boolean',
                'description': 'Limit Elasticsearch to index certain namespaces and projects',
            },
            'elasticsearch_project_ids': {  # >= PREMIUM, SILVER
                'type': 'array',
                'description': 'The projects to index via Elasticsearch if elasticsearch_limit_indexing is enabled.',
                'items': {
                    'type': 'integer',
                },
                'uniqueItems': True,
            },
            'elasticsearch_namespace_ids': {  # >= PREMIUM, SILVER
                'type': 'array',
                'description': 'The namespaces to index via Elasticsearch if elasticsearch_limit_indexing is enabled.',
                'items': {
                    'type': 'integer',
                },
                'uniqueItems': True,
            },
            'email_additional_text': {  # >= PREMIUM, SILVER
                'type': 'string',
                'description': 'Additional text added to the bottom of every email for legal/auditing/compliance '
                               'reasons',
            },
            'email_author_in_body': {
                'type': 'boolean',
                'description': 'Some email servers do not support overriding the email sender name. Enable this option '
                               'to include the name of the author of the issue, merge request or comment in the email '
                               'body instead.',
            },
            'enabled_git_access_protocol': {
                'type': 'string',
                'description': 'Enabled protocols for Git access. Allowed values are: ssh, http, and nil to allow both '
                               'protocols.',
                'enum': ['ssh', 'http', 'both'],
            },
            'enforce_terms': {
                'type': 'boolean',
                'description': 'Enforce application ToS to all users.',
            },
            'external_auth_client_cert': {
                'type': 'string',
                'description': 'The certificate to use to authenticate with the external authorization service',
            },
            'external_auth_client_key': {
                'type': 'string',
                'description': 'Private key for the certificate when authentication is required for the external '
                               'authorization service, this is encrypted when stored',
            },
            'external_auth_client_key_pass': {
                'type': 'string',
                'description': 'Passphrase to use for the private key when authenticating with the external service '
                               'this is encrypted when stored',
            },
            'external_authorization_service_default_label': {
                'type': 'string',
                'description': 'The default classification label to use when requesting authorization and no '
                               'classification label has been specified on the project',
            },
            'external_authorization_service_enabled': {
                'type': 'boolean',
                'description': 'Enable using an external authorization service for accessing projects',
            },
            'external_authorization_service_timeout': {
                'type': 'number',
                'description': 'The timeout after which an authorization request is aborted, in seconds. When a '
                               'request times out, access is denied to the user. (min: 0.001, max: 10, step: 0.001)',
            },
            'external_authorization_service_url': {
                'type': 'string',
                'description': 'URL to which authorization requests will be directed',
            },
            'file_template_project_id': {  # >= PREMIUM, SILVER
                'type': 'integer',
                'description': 'The ID of a project to load custom file templates from',
            },
            'first_day_of_week': {
                'type': 'string',
                'description': 'Start day of the week for calendar views and date pickers.',
                'enum': ['sunday', 'monday', 'saturday'],
            },
            'geo_node_allowed_ips': {  # PREMIUM, SILVER
                'type': 'string',
                'description': 'Comma-separated list of IPs and CIDRs of allowed secondary nodes. For example, '
                               '1.1.1.1, 2.2.2.0/24.',
            },
            'geo_status_timeout': {  # >= PREMIUM, SILVER
                'type': 'integer',
                'description': 'The amount of seconds after which a request to get a secondary node status will time '
                               'out.',
            },
            'gitaly_timeout_default': {
                'type': 'integer',
                'description': 'Default Gitaly timeout, in seconds. This timeout is not enforced for git fetch/push '
                               'operations or Sidekiq jobs. Set to 0 to disable timeouts.',
            },
            'gitaly_timeout_fast': {
                'type': 'integer',
                'description': 'Gitaly fast operation timeout, in seconds. Some Gitaly operations are expected to be '
                               'fast. If they exceed this threshold, there may be a problem with a storage shard and '
                               '‘failing fast’ can help maintain the stability of the GitLab instance. Set to 0 to '
                               'disable timeouts.',
            },
            'gitaly_timeout_medium': {
                'type': 'integer',
                'description': 'Medium Gitaly timeout, in seconds. This should be a value between the Fast and the '
                               'Default timeout. Set to 0 to disable timeouts.',
            },
            'grafana_enabled': {
                'type': 'boolean',
                'description': 'Enable Grafana.',
            },
            'grafana_url': {
                'type': 'string',
                'description': 'Grafana URL.',
            },
            'gravatar_enabled': {
                'type': 'boolean',
                'description': 'Enable Gravatar.',
            },
            'hashed_storage_enabled': {
                'type': 'boolean',
                'description': 'Create new projects using hashed storage paths: Enable immutable, hash-based paths and '
                               'repository names to store repositories on disk. This prevents repositories from having '
                               'to be moved or renamed when the Project URL changes and may improve disk I/O '
                               'performance. (EXPERIMENTAL)',
            },
            'help_page_hide_commercial_content': {
                'type': 'boolean',
                'description': 'Hide marketing-related entries from help.',
            },
            'help_page_support_url': {
                'type': 'string',
                'description': 'Alternate support URL for help page.',
            },
            'help_page_text': {
                'type': 'string',
                'description': 'Custom text displayed on the help page.',
            },
            'help_text': {  # >= PREMIUM, SILVER
                'type': 'string',
                'description': 'GitLab server administrator information',
            },
            'hide_third_party_offers': {
                'type': 'boolean',
                'description': 'Do not display offers from third parties within GitLab.',
            },
            'home_page_url': {
                'type': 'string',
                'description': 'Redirect to this URL when not logged in.',
            },
            'housekeeping_bitmaps_enabled': {
                'type': 'boolean',
                'description': 'Enable Git pack file bitmap creation.',
            },
            'housekeeping_enabled': {
                'type': 'boolean',
                'description': 'Enable or disable git housekeeping.',
            },
            'housekeeping_full_repack_period': {
                'type': 'integer',
                'description': 'Number of Git pushes after which an incremental git repack is run.',
            },
            'housekeeping_gc_period': {
                'type': 'integer',
                'description': 'Number of Git pushes after which git gc is run.',
            },
            'housekeeping_incremental_repack_period': {
                'type': 'integer',
                'description': 'Number of Git pushes after which an incremental git repack is run.',
            },
            'html_emails_enabled': {
                'type': 'boolean',
                'description': 'Enable HTML emails.',
            },
            'import_sources': {
                'type': 'array',
                'description': 'Sources to allow project import from, possible values.',
                'items': {
                    'type': 'string',
                    'enum': [
                        'github',
                        'bitbucket',
                        'bitbucket_server',
                        'gitlab',
                        'google_code',
                        'fogbugz',
                        'git',
                        'gitlab_project',
                        'gitea',
                        'manifest',
                        'phabricator',
                    ],
                },
                'uniqueItems': True,
            },
            'instance_statistics_visibility_private': {
                'type': 'boolean',
                'description': 'When set to true Instance statistics will only be available to admins.',
            },
            'local_markdown_version': {
                'type': 'integer',
                'description': 'Increase this value when any cached markdown should be invalidated.',
            },
            'max_artifacts_size': {
                'type': 'integer',
                'description': 'Maximum artifacts size in MB',
            },
            'max_attachment_size': {
                'type': 'integer',
                'description': 'Limit attachment size in MB',
            },
            'max_pages_size': {
                'type': 'integer',
                'description': 'Maximum size of pages repositories in MB',
            },
            'metrics_enabled': {
                'type': 'boolean',
                'description': 'Enable influxDB metrics.',
            },
            'metrics_host': {
                'type': 'string',
                'description': 'InfluxDB host.',
            },
            'metrics_method_call_threshold': {
                'type': 'integer',
                'description': 'A method call is only tracked when it takes longer than the given amount of '
                               'milliseconds.',
            },
            'metrics_packet_size': {
                'type': 'integer',
                'description': 'The amount of datapoints to send in a single UDP packet.',
            },
            'metrics_pool_size': {
                'type': 'integer',
                'description': 'The amount of InfluxDB connections to keep open.',
            },
            'metrics_port': {
                'type': 'integer',
                'description': 'The UDP port to use for connecting to InfluxDB.',
            },
            'metrics_sample_interval': {
                'type': 'integer',
                'description': 'The sampling interval in seconds.',
            },
            'metrics_timeout': {
                'type': 'integer',
                'description': 'The amount of seconds after which InfluxDB will time out.',
            },
            'mirror_available': {
                'type': 'boolean',
                'description': 'Allow mirrors to be set up for projects. If disabled, only admins will be able to set '
                               'up mirrors in projects.',
            },
            'mirror_capacity_threshold': {  # >= PREMIUM, SILVER
                'type': 'integer',
                'description': 'Minimum capacity to be available before scheduling more mirrors preemptively',
            },
            'mirror_max_capacity': {  # >= PREMIUM, SILVER
                'type': 'string',
                'description': 'Maximum number of mirrors that can be synchronizing at the same time.',
            },
            'mirror_max_delay': {  # >= PREMIUM, SILVER
                'type': 'integer',
                'description': 'Maximum time (in minutes) between updates that a mirror can have when scheduled to '
                               'synchronize.',
            },
            'pages_domain_verification_enabled': {
                'type': 'boolean',
                'description': 'Require users to prove ownership of custom domains. Domain verification is an '
                               'essential security measure for public GitLab sites. Users are required to demonstrate '
                               'they control a domain before it is enabled.',
            },
            'password_authentication_enabled_for_git': {
                'type': 'boolean',
                'description': 'Enable authentication for Git over HTTP(S) via a GitLab account password. Default is '
                               'true.',
            },
            'password_authentication_enabled_for_web': {
                'type': 'boolean',
                'description': 'Enable authentication for the web interface via a GitLab account password. Default is '
                               'true.',
            },
            'performance_bar_allowed_group_path': {
                'type': 'string',
                'description': 'Path of the group that is allowed to toggle the performance bar.',
            },
            'plantuml_enabled': {
                'type': 'boolean',
                'description': 'Enable PlantUML integration.',
            },
            'plantuml_url': {
                'type': 'string',
                'description': 'The PlantUML instance URL for integration.',
            },
            'polling_interval_multiplier': {
                'type': 'number',
                'description': 'Interval multiplier used by endpoints that perform polling. Set to 0 to disable '
                               'polling.',
            },
            'project_export_enabled': {
                'type': 'boolean',
                'description': 'Enable project export.',
            },
            'prometheus_metrics_enabled': {
                'type': 'boolean',
                'description': 'Enable prometheus metrics.',
            },
            'protected_ci_variables': {
                'type': 'boolean',
                'description': 'Environment variables are protected by default.',
            },
            'pseudonymizer_enabled': {  # >= PREMIUM, SILVER
                'type': 'boolean',
                'description': 'When enabled, GitLab will run a background job that will produce pseudonymized CSVs of '
                               'the GitLab database that will be uploaded to your configured object storage directory.',
            },
            'recaptcha_enabled': {
                'type': 'boolean',
                'description': 'Enable recaptcha.',
            },
            'recaptcha_private_key': {
                'type': 'string',
                'description': 'Private key for recaptcha.',
            },
            'recaptcha_site_key': {
                'type': 'string',
                'description': 'Site key for recaptcha.',
            },
            'receive_max_input_size': {
                'type': 'integer',
                'description': 'Maximum push size (MB).',
            },
            'repository_checks_enabled': {
                'type': 'boolean',
                'description': 'GitLab will periodically run git fsck in all project and wiki repositories to look for '
                               'silent disk corruption issues.',
            },
            'repository_size_limit': {  # >= PREMIUM, SILVER
                'type': 'string',
                'description': 'Size limit per repository (MB)',
            },
            'repository_storages': {
                'type': 'array',
                'description': 'A list of names of enabled storage paths, taken from gitlab.yml. New projects will be '
                               'created in one of these stores, chosen at random.',
                'items': {
                    'type': 'string',
                },
                'uniqueItems': True,
            },
            'require_two_factor_authentication': {
                'type': 'boolean',
                'description': 'Require all users to set up Two-factor authentication.',
            },
            'restricted_visibility_levels': {
                'type': 'array',
                'description': 'Selected levels cannot be used by non-admin users for groups, projects or snippets. '
                               'Can take private, internal and public as a parameter. Default is null which means '
                               'there is no restriction.',
                'items': {
                    'type': 'string',
                    'enum': ['private', 'internal', 'public'],
                },
                'uniqueItems': True,
            },
            'rsa_key_restriction': {
                'type': 'integer',
                'description': 'The minimum allowed bit length of an uploaded RSA key. Default is 0 (no restriction). '
                               '-1 disables RSA keys.',
            },
            'send_user_confirmation_email': {
                'type': 'boolean',
                'description': 'Send confirmation email on sign-up.',
            },
            'session_expire_delay': {
                'type': 'integer',
                'description': 'Session duration in minutes. GitLab restart is required to apply changes',
            },
            'shared_runners_enabled': {
                'type': 'boolean',
                'description': 'Enable shared runners for new projects.',
            },
            'shared_runners_minutes': {  # PREMIUM, SILVER
                'type': 'integer',
                'description': 'Set the maximum number of pipeline minutes that a group can use on shared Runners per '
                               'month.',
            },
            'shared_runners_text': {
                'type': 'string',
                'description': 'Shared runners text.',
            },
            'sign_in_text': {
                'type': 'string',
                'description': 'Text on the login page.',
            },
            'signup_enabled': {
                'type': 'boolean',
                'description': 'Enable registration.',
            },
            'slack_app_enabled': {  # PREMIUM, SILVER
                'type': 'boolean',
                'description': ' Enable Slack app.',
            },
            'slack_app_id': {  # PREMIUM, SILVER
                'type': 'string',
                'description': 'The app id of the Slack-app.',
            },
            'slack_app_secret': {  # PREMIUM, SILVER
                'type': 'string',
                'description': 'The app secret of the Slack-app.',
            },
            'slack_app_verification_token': {  # PREMIUM, SILVER
                'type': 'string',
                'description': 'The verification token of the Slack-app.',
            },
            'terminal_max_session_time': {
                'type': 'integer',
                'description': 'Maximum time for web terminal websocket connection (in seconds). Set to 0 for '
                               'unlimited time.',
            },
            'terms': {
                'type': 'string',
                'description': 'Markdown content for the ToS.',
            },
            'throttle_authenticated_api_enabled': {
                'type': 'boolean',
                'description': 'Enable authenticated API request rate limit. Helps reduce request volume (e.g. from '
                               'crawlers or abusive bots).',
            },
            'throttle_authenticated_api_period_in_seconds': {
                'type': 'integer',
                'description': 'Rate limit period in seconds.',
            },
            'throttle_authenticated_api_requests_per_period': {
                'type': 'integer',
                'description': 'Max requests per period per user.',
            },
            'throttle_authenticated_web_enabled': {
                'type': 'boolean',
                'description': 'Enable authenticated web request rate limit. Helps reduce request volume (e.g. from '
                               'crawlers or abusive bots).',
            },
            'throttle_authenticated_web_period_in_seconds': {
                'type': 'integer',
                'description': 'Rate limit period in seconds.',
            },
            'throttle_authenticated_web_requests_per_period': {
                'type': 'integer',
                'description': 'Max requests per period per user.',
            },
            'throttle_unauthenticated_enabled': {
                'type': 'boolean',
                'description': 'Enable unauthenticated request rate limit. Helps reduce request volume (e.g. from '
                               'crawlers or abusive bots).',
            },
            'throttle_unauthenticated_period_in_seconds': {
                'type': 'integer',
                'description': 'Rate limit period in seconds.',
            },
            'throttle_unauthenticated_requests_per_period': {
                'type': 'integer',
                'description': 'Max requests per period per IP.',
            },
            'time_tracking_limit_to_hours': {
                'type': 'boolean',
                'description': 'Limit display of time tracking units to hours. Default is false.',
            },
            'two_factor_grace_period': {
                'type': 'integer',
                'description': 'Amount of time (in hours) that users are allowed to skip forced configuration of '
                               'two-factor authentication.',
            },
            'unique_ips_limit_enabled': {
                'type': 'boolean',
                'description': 'Limit sign in from multiple ips.',
            },
            'unique_ips_limit_per_user': {
                'type': 'integer',
                'description': 'Maximum number of ips per user.',
            },
            'unique_ips_limit_time_window': {
                'type': 'integer',
                'description': 'How many seconds an IP will be counted towards the limit.',
            },
            'usage_ping_enabled': {
                'type': 'boolean',
                'description': 'Every week GitLab will report license usage back to GitLab, Inc.',
            },
            'user_default_external': {
                'type': 'boolean',
                'description': 'Newly registered users will be external by default.',
            },
            'user_default_internal_regex': {
                'type': 'string',
                'description': 'Specify an e-mail address regex pattern to identify default internal users.',
            },
            'user_oauth_applications': {
                'type': 'boolean',
                'description': 'Allow users to register any application to use GitLab as an OAuth provider.',
            },
            'user_show_add_ssh_key_message': {
                'type': 'boolean',
                'description': 'When set to false disable the “You won’t be able to pull or push project code via SSH” '
                               'warning shown to users with no uploaded SSH key.',
            },
            'version_check_enabled': {
                'type': 'boolean',
                'description': 'Let GitLab inform you when an update is available.',
            },
            'web_ide_clientside_preview_enabled': {
                'type': 'boolean',
                'description': 'Client side evaluation (Allow live previews of JavaScript projects in the Web IDE '
                               'using CodeSandbox client side evaluation).',
            },
        },
        'additionalProperties': False,
        'dependencies': {
            'akismet_enabled': ['akismet_api_key'],
            'domain_blacklist_enabled': ['domain_blacklist'],
            'enforce_terms': ['terms'],
            'external_auth_client_cert': ['external_auth_client_key'],
            'external_authorization_service_enabled': ['external_authorization_service_timeout',
                                                       'external_authorization_service_url'],
            'housekeeping_enabled': ['housekeeping_bitmaps_enabled', 'housekeeping_full_repack_period',
                                     'housekeeping_gc_period', 'housekeeping_incremental_repack_period'],
            'metrics_enabled': ['metrics_host', 'metrics_method_call_threshold', 'metrics_packet_size',
                                'metrics_pool_size', 'metrics_port', 'metrics_sample_interval', 'metrics_timeout'],
            'plantuml_enabled': ['plantuml_url'],
            'recaptcha_enabled': ['recaptcha_site_key', 'recaptcha_private_key'],
            'require_two_factor_authentication': ['two_factor_grace_period'],
            'shared_runners_enabled': ['shared_runners_text'],
        },
    }

    """"_object_manager()

    Return the python-gitlab Gilab object.
    """
    def _object_manager(self):
        return self.pygitlab.settings

    """"_get()

    Set the _object attribute
    """
    def _get(self):
        obj_manager = self._object_manager()
        self._obj = obj_manager.get()

    """"_mangle_param()

    Convert a param value from GitLabracadabra form to API form.
    """
    def _mangle_param(self, param_name, param_value):
        if param_name == 'first_day_of_week':
            if param_value == 'sunday':
                return 0
            elif param_value == 'monday':
                return 1
            elif param_value == 'saturday':
                return 6
            else:
                return param_value
        else:
            return GitLabracadabraObject._mangle_param(self, param_name, param_value)

    """"_unmangle_param()

    Convert a param value from API form to GitLabracadabra form.
    """
    def _unmangle_param(self, param_name, param_value):
        if param_name == 'first_day_of_week':
            if param_value == 0:
                return 'sunday'
            elif param_value == 1:
                return 'monday'
            elif param_value == 6:
                return 'saturday'
            else:
                return param_value
        else:
            return GitLabracadabraObject._unmangle_param(self, param_name, param_value)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013-2017 Gauvain Pocentek <gauvain@pocentek.net>
# Copyright (C) 2019-2021 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re

import yaml

from gitlabracadabra.dictutils import update_dict_with_defaults


class GitlabracadabraParser(object):
    """YAML parser."""

    def __init__(self, action_file, config, recursion=0):
        self._action_file = action_file
        self._config = config
        self._objects = None
        self._include = self._config.pop('include', [])
        for included in self._include:
            if recursion >= 10:
                raise ValueError('%s: nesting too deep in `include`' % self._action_file)
            if isinstance(included, str):
                included = {'local': included}
            if not isinstance(included, dict):
                raise ValueError('%s: invalid value for `include`: %s' % (self._action_file, included))
            if list(included.keys()) == ['local'] and isinstance(included['local'], str):
                if '..' in included['local'] or included['local'][0] == '/':
                    raise ValueError('%s: forbidden path for `include`: %s' % (self._action_file, included['local']))
                included = self.from_yaml_file(included['local'], recursion + 1)
            else:
                raise ValueError('%s: invalid value for `include`: %s' % (self._action_file, included))
            update_dict_with_defaults(self._config, included._config)

    @classmethod
    def from_yaml(cls, action_file, yaml_blob, recursion=0):
        config = yaml.safe_load(yaml_blob)
        return GitlabracadabraParser(action_file, config, recursion)

    @classmethod
    def from_yaml_file(cls, action_file, recursion=0):
        with open(action_file) as yaml_blob:
            return cls.from_yaml(action_file, yaml_blob, recursion)

    """items()

    Handle hidden objects (starting with a dot) and extends.
    """
    def _items(self):
        for k, v in sorted(self._config.items()):
            if k.startswith('.'):
                continue
            recursion = 0
            while 'extends' in v:
                recursion += 1
                if recursion >= 10:
                    raise ValueError('%s (%s): nesting too deep in `extends`' % (self._action_file, k))
                # No need to deepcopy as update_dict_with_defaults() does
                v = v.copy()
                extends = v.pop('extends')
                if isinstance(extends, str):
                    extends = [extends]
                for extends_item in reversed(extends):
                    if isinstance(extends_item, str):
                        extends_item = {extends_item: 'deep'}
                    for extends_k, extends_v in extends_item.items():
                        try:
                            parent = self._config[extends_k]
                        except KeyError:
                            raise ValueError('%s (`%s` from `%s`): %s not found' %
                                             (self._action_file, extends_k, k, extends_k))
                        if extends_v == 'deep':
                            update_dict_with_defaults(v, parent)
                        elif extends_v == 'replace':
                            result = parent.copy()
                            result.update(v)
                            v = result
                        elif extends_v == 'aggregate':
                            update_dict_with_defaults(v, parent, aggregate=True)
                        else:
                            raise ValueError('%s (`%s` from `%s`): Unknown merge strategy `%s`' %
                                             (self._action_file, extends_k, k, extends_v))
            # Drop None values from v
            yield (k, {a: b for a, b in v.items() if b is not None})

    """_type_to_classname()

    Converts object-type to GitLabracadabraObjectType.
    """
    @classmethod
    def _type_to_classname(cls, obj_type):
        splitted = re.split('[-_]', obj_type)
        mapped = map(lambda s: s[0].upper() + s[1:].lower(), splitted)
        return 'GitLabracadabra' + ''.join(mapped)

    """_type_to_module()

    Converts object-type to gitlabracadabra.objects.object_type.
    """
    @classmethod
    def _type_to_module(cls, obj_type):
        return 'gitlabracadabra.objects.' + obj_type.lower().replace('-', '_')

    """get_class_for()

    Get the class for the given object type.
    """
    @classmethod
    def get_class_for(cls, obj_type):
        obj_classname = cls._type_to_classname(obj_type)
        obj_module = __import__(cls._type_to_module(obj_type), globals(), locals(), [obj_classname])
        return getattr(obj_module, obj_classname)

    """objects()

    Returns .
    """
    def objects(self):
        if self._objects is not None:
            return self._objects
        self._objects = {}
        for k, v in self._items():
            if 'type' in v:
                obj_type = v['type']
                v.pop('type')
            elif k.endswith('/'):
                obj_type = 'group'
            else:
                obj_type = 'project'
            if k.endswith('/'):
                k = k[:-1]
            obj_class = self.get_class_for(obj_type)
            self._objects[k] = obj_class(self._action_file, k, v)
        return self._objects



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import logging
import re
from copy import deepcopy
from typing import TYPE_CHECKING, List, Optional

from gitlab.exceptions import GitlabCreateError, GitlabDeleteError, GitlabGetError, GitlabListError, GitlabUpdateError

from jsonschema.validators import validator_for

from gitlabracadabra.gitlab.connections import GitlabConnections


if TYPE_CHECKING:
    from pygit2 import UserPass

    from gitlab import Gitlab

    from gitlabracadabra.auth_info import AuthInfo
    from gitlabracadabra.gitlab.connection import GitlabConnection

logger = logging.getLogger(__name__)


class GitLabracadabraObject(object):
    SCHEMA = {
        '$schema': 'http://json-schema.org/draft-04/schema#',
        'title': 'Object',
        'type': 'object',
        'properties': {
        },
        'additionalProperties': False,
        # 'required': [],
    }

    # If not None, use find(FIND_PARAM=...) instead of get(...)
    FIND_PARAM: Optional[str] = None

    # If not None, set to id of the object on create
    CREATE_KEY: Optional[str] = None

    CREATE_PARAMS: List[str] = []

    IGNORED_PARAMS: List[str] = []

    def __init__(self, action_file, name, content):
        self._action_file = action_file
        self._name = name
        self._content = content
        validator_class = validator_for(self.SCHEMA)
        validator_class.check_schema(self.SCHEMA)
        validator = validator_class(self.SCHEMA)
        self._errors = sorted(validator.iter_errors(content), key=lambda e: e.path)
        self._gitlab_id = self._content.pop('gitlab_id', None)
        self._create_object = self._content.pop('create_object', None)
        self._delete_object = self._content.pop('delete_object', None)

    @property
    def connection(self) -> GitlabConnection:
        return GitlabConnections().get_connection(self._gitlab_id)

    @property
    def pygitlab(self) -> Gitlab:
        return self.connection.pygitlab

    @property
    def pygit2_credentials(self) -> UserPass:
        return self.connection.pygit2_credentials

    @property
    def registry_auth_info(self) -> AuthInfo:
        return self.connection.registry_auth_info

    def errors(self):
        return self._errors

    """"type_name()

    GitLabracadabraProject -> project.
    """
    @classmethod
    def _type_name(cls):
        return cls.__name__[15:].lower()

    """"type_name_plural()

    GitLabracadabraProject -> projects.
    """
    @classmethod
    def _type_name_plural(cls):
        return cls._type_name() + 's'

    """"_object_manager()

    Return the python-gitlab Gilab object.
    """
    def _object_manager(self):
        return getattr(self.pygitlab, self._type_name_plural())

    """"_create()

    Create the object.
    """
    def _create(self, dry_run=False):
        obj_manager = self._object_manager()
        namespace_manager = self.pygitlab.namespaces
        namespaces = self._name.split('/')
        object_path = namespaces.pop()
        create_params = {
            'path': object_path,
        }
        if self.CREATE_KEY:
            create_params[self.CREATE_KEY] = object_path
        for param_name in self.CREATE_PARAMS:
            if param_name in self._content:
                create_params[param_name] = self._content[param_name]
        if len(namespaces):
            try:
                parent_namespace = namespace_manager.get('/'.join(namespaces))
            except GitlabGetError as e:
                error_message = e.error_message
                if e.response_code == 404:
                    error_message = 'parent namespace not found'
                logger.error('[%s] NOT Creating %s (%s)', self._name, self._type_name(), error_message)
                return None
            if self._type_name() == 'group':
                create_params['parent_id'] = parent_namespace.id
            else:
                create_params['namespace_id'] = parent_namespace.id
        if dry_run:
            logger.info('[%s] NOT Creating %s (dry-run)', self._name, self._type_name())
            return None
        else:
            logger.info('[%s] Creating %s', self._name, self._type_name())
            try:
                return obj_manager.create(create_params)
            except GitlabCreateError as e:
                logger.error('[%s] NOT Creating %s (%s)', self._name, self._type_name(), e.error_message)
                return None

    """"_delete()

    Delete the object.
    """
    def _delete(self, dry_run=False):
        if self._obj is None:
            logger.debug('[%s] NOT Deleting %s (not found)', self._name, self._type_name())
        elif dry_run:
            logger.info('[%s] NOT Deleting %s (dry-run)', self._name, self._type_name())
        else:
            logger.info('[%s] Deleting %s', self._name, self._type_name())
            try:
                self._obj.delete()
            except GitlabCreateError as e:
                logger.error('[%s] Unable to delete %s (%s)', self._name, self._type_name(), e.error_message)

    """"_mangle_param()

    Convert a param value from GitLabracadabra form to API form.
    """
    def _mangle_param(self, param_name, param_value):
        if isinstance(param_value, str):
            # GitLab normalize to CRLF
            # YAML normalize to LF
            return param_value.replace('\n', '\r\n')
        return param_value

    """"_unmangle_param()

    Convert a param value from API form to GitLabracadabra form.
    """
    def _unmangle_param(self, param_name, param_value):
        return param_value

    """"_canonalize_param()

    Canonalize a param value.
    """
    def _canonalize_param(self, param_name, param_value):
        if isinstance(param_value, list):
            return sorted(param_value)
        return param_value

    """"_get_param()

    Get a param value.
    """
    def _get_param(self, param_name):
        return getattr(self._obj, param_name)

    """"_process_param()

    Process one param.
    """
    def _process_param(self, param_name, param_value, dry_run=False, skip_save=False):
        if param_name in self.IGNORED_PARAMS and not skip_save:
            return
        target_value = self._canonalize_param(param_name, param_value)
        try:
            current_value = self._canonalize_param(param_name,
                                                   self._unmangle_param(param_name, self._get_param(param_name)))
        except AttributeError:
            if not skip_save:
                # FIXME Hidden attributes cannot be idempotent (i.e password)
                logger.info('[%s] NOT Changing param %s: %s -> %s (current value is not available)',
                            self._name, param_name, None, target_value)
                return
            current_value = None
        if current_value != target_value:
            if 'dependencies' in self.SCHEMA and param_name in self.SCHEMA['dependencies']:
                for dependency in self.SCHEMA['dependencies'][param_name]:
                    process_method = getattr(self, '_process_' + dependency, self._process_param)
                    process_method(dependency, self._content[dependency], dry_run, skip_save=True)
            if dry_run:
                logger.info('[%s] NOT Changing param %s: %s -> %s (dry-run)',
                            self._name, param_name, current_value, target_value)
                setattr(self._obj, param_name, self._mangle_param(param_name, target_value))
            else:
                logger.info('[%s] Changing param %s: %s -> %s',
                            self._name, param_name, current_value, target_value)
                setattr(self._obj, param_name, target_value)
                if not skip_save:
                    try:
                        self._obj.save()
                    except GitlabUpdateError as e:
                        logger.error('[%s] Unable to change param %s (%s -> %s): %s',
                                     self._name, param_name, current_value, target_value, e.error_message)

    """"_get()

    Set the _object attribute
    """
    def _get(self):
        obj_manager = self._object_manager()
        if self.FIND_PARAM:
            params = {self.FIND_PARAM: self._name}
            try:
                self._obj = obj_manager.list(**params)[0]
            except IndexError:
                self._obj = None
        else:
            try:
                self._obj = obj_manager.get(self._name)
            except GitlabGetError as err:
                if err.response_code != 404:
                    pass
                self._obj = None

    """"process()

    Process the object.
    """
    def process(self, dry_run=False):
        content_copy = deepcopy(self._content)
        self._get()
        if self._delete_object:
            self._delete(dry_run)
            return
        if self._obj is None:
            if self._create_object:
                self._obj = self._create(dry_run)
                if self._obj is None:
                    return
            else:
                logger.info('[%s] NOT Creating %s (create_object is false)', self._name, self._type_name())
                return
        for param_name, param_value in sorted(self._content.items()):
            process_method = getattr(self, '_process_' + param_name, self._process_param)
            try:
                process_method(param_name, param_value, dry_run)
            except (GitlabCreateError, GitlabDeleteError, GitlabGetError, GitlabListError, GitlabUpdateError) as e:
                logger.error('[%s] Error while processing param %s: %s',
                             self._name, param_name, e.error_message)
        if content_copy != self._content:
            raise RuntimeError('[%s] Changed values during processing' % self._name)

    """"_markdown_link()

    Generate a Markdown link.
    """
    @classmethod
    def _markdown_link(cls, header):
        # Remove heading #
        out = re.sub(r'^#+', '', header)
        # Trim and lower
        out = out.strip().lower()
        # Remove non-word characters
        out = re.sub(r'[^\w\- ]+', ' ', out)
        # Replace multiple spaces with dash
        out = re.sub(r'\s+', '-', out)
        # Remove trailing dashed
        out = re.sub(r'-+$', '', out)
        return '#' + out

    """"doc_markdown()

    Generate Markdown documentation.
    """
    @classmethod
    def doc_markdown(cls):
        output = ''
        properties = cls.SCHEMA.get('properties', {})
        first_undocumented = True
        for p in sorted(properties):
            if p not in cls.DOC:
                if first_undocumented:
                    cls.DOC.append('# Undocumented')
                    first_undocumented = False
                cls.DOC.append(p)
        # Generate TOC
        output += '# Table of Contents <!-- omit in toc -->' + '\n\n'
        for doc in cls.DOC:
            if doc[0] == '#':
                matches = re.match(r'^(#+) ?(.*)$', doc)
                indent = '  ' * (len(matches.group(1)) - 1)
                output += indent + '- [' + doc.replace('#', '').strip() + '](' + cls._markdown_link(doc) + ')\n'
        output += '\n'
        # Detail
        for doc in cls.DOC:
            if doc[0] == '#':
                output += doc + '\n\n'
            elif doc in properties:
                example = cls.EXAMPLE_YAML_HEADER
                if '_example' in properties[doc]:
                    if properties[doc]['_example'].startswith('\n'):
                        example += '  ' + doc + ':' + properties[doc]['_example'] + '\n'
                    else:
                        example += '  ' + doc + ': ' + properties[doc]['_example'] + '\n'
                elif 'enum' in properties[doc]:
                    enum = properties[doc]['enum']
                    example += ('  ' + doc + ': ' + str(enum[0]) +
                                ' # one of ' + ', '.join([str(item) for item in enum]) + '\n')
                elif properties[doc]['type'] == 'boolean':
                    example += '  ' + doc + ': true # or false\n'
                elif properties[doc]['type'] == 'integer':
                    example += '  ' + doc + ': 42\n'
                elif properties[doc]['type'] == 'array':
                    example += '  ' + doc + ': [My ' + doc.replace('_', ' ') + ']\n'
                else:
                    example += '  ' + doc + ': My ' + doc.replace('_', ' ') + '\n'
                output += '`' + doc + '` ' + properties[doc].get('description', 'Undocumented') + ':\n'
                output += '```yaml\n' + example + '```\n\n'
                if '_doc_link' in properties[doc]:
                    output += 'More information can be found [here]({0})\n\n'.format(properties[doc]['_doc_link'])
            else:
                output += '`' + doc + '` *Undocumented*.\n\n'
        return output



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from copy import deepcopy


logger = logging.getLogger(__name__)


class PipelineSchedulesMixin(object):
    """Object with pipeline schedules."""

    """_process_pipeline_schedules()

    Process the pipeline_schedules param.
    """
    def _process_pipeline_schedules(self, param_name, param_value, dry_run=False, skip_save=False):
        assert param_name == 'pipeline_schedules'  # noqa: S101
        assert not skip_save  # noqa: S101
        unknown_pipeline_schedules = self._content.get('unknown_pipeline_schedules', 'warn')
        current_pipeline_schedules = dict([[current_pipeline_schedule.description, current_pipeline_schedule]
                                          for current_pipeline_schedule in self._obj.pipelineschedules.list(all=True)])
        target_pipeline_schedules = dict([[target_pipeline_schedule['description'], deepcopy(target_pipeline_schedule)]
                                         for target_pipeline_schedule in param_value])
        unknown_pipeline_schedule_variables = self._content.get('unknown_pipeline_schedule_variables', 'warn')
        # We first check for already existing pipeline schedules
        for current_pipeline_schedule_description, current_pipeline_schedule in (
            sorted(current_pipeline_schedules.items())
        ):
            if current_pipeline_schedule_description in target_pipeline_schedules:
                target_pipeline_schedule = target_pipeline_schedules[current_pipeline_schedule_description]
                for target_pipeline_schedule_param_name, target_pipeline_schedule_param_value in (
                    target_pipeline_schedule.items()
                ):
                    if target_pipeline_schedule_param_name == 'variables':
                        unknown_variables = target_pipeline_schedule.get('unknown_variables',
                                                                         unknown_pipeline_schedule_variables)
                        self._handle_pipeline_schedule_variables(current_pipeline_schedule,
                                                                 target_pipeline_schedule_param_value,
                                                                 unknown_variables,
                                                                 dry_run)
                        continue
                    if target_pipeline_schedule_param_name == 'unknown_variables':
                        continue
                    try:
                        current_pipeline_schedule_param_value = getattr(current_pipeline_schedule,
                                                                        target_pipeline_schedule_param_name)
                    except AttributeError:
                        logger.info('[%s] NOT Changing pipeline schedules %s %s: %s -> %s '
                                    '(current value is not available)',
                                    self._name, current_pipeline_schedule_description,
                                    target_pipeline_schedule_param_name, None, target_pipeline_schedule_param_value)
                        continue
                    if current_pipeline_schedule_param_value != target_pipeline_schedule_param_value:
                        if dry_run:
                            logger.info('[%s] NOT Changing pipeline schedule %s %s: %s -> %s (dry-run)',
                                        self._name, current_pipeline_schedule_description,
                                        target_pipeline_schedule_param_name, current_pipeline_schedule_param_value,
                                        target_pipeline_schedule_param_value)
                        else:
                            logger.info('[%s] Changing pipeline schedule %s %s: %s -> %s',
                                        self._name, current_pipeline_schedule_description,
                                        target_pipeline_schedule_param_name, current_pipeline_schedule_param_value,
                                        target_pipeline_schedule_param_value)
                            setattr(current_pipeline_schedule, target_pipeline_schedule_param_name,
                                    target_pipeline_schedule_param_value)
                            current_pipeline_schedule.save()
                target_pipeline_schedules.pop(current_pipeline_schedule_description)
            else:
                if unknown_pipeline_schedules in ['delete', 'remove']:
                    if dry_run:
                        logger.info('[%s] NOT Removing pipeline schedule %s (dry-run)',
                                    self._name, current_pipeline_schedule_description)
                    else:
                        logger.info('[%s] Removing pipeline schedule %s',
                                    self._name, current_pipeline_schedule_description)
                        current_pipeline_schedule.delete()
                elif unknown_pipeline_schedules not in ['ignore', 'skip']:
                    logger.warning('[%s] NOT Removing pipeline schedule: %s',
                                   self._name, current_pipeline_schedule_description)
        # Remaining pipeline_schedules
        for target_pipeline_schedule_name, target_pipeline_schedule in sorted(target_pipeline_schedules.items()):
            target_pipeline_schedule_variables = target_pipeline_schedule.pop('variables', [])
            unknown_variables = target_pipeline_schedule.pop('unknown_variables', unknown_pipeline_schedule_variables)
            if dry_run:
                logger.info('[%s] NOT Adding pipeline schedule %s: %s -> %s (dry-run)',
                            self._name, target_pipeline_schedule_name, None, target_pipeline_schedule)
            else:
                logger.info('[%s] Adding pipeline schedule %s: %s -> %s',
                            self._name, target_pipeline_schedule_name, None, target_pipeline_schedule)
                pipeline_schedule = self._obj.pipelineschedules.create(target_pipeline_schedule)
                self._handle_pipeline_schedule_variables(pipeline_schedule, target_pipeline_schedule_variables,
                                                         unknown_variables, dry_run)

    """_handle_pipeline_schedule_variables()

    Handle pipeline schedule variables.
    """
    def _handle_pipeline_schedule_variables(self, pipeline_schedule, variables, unknown_variables, dry_run=False):
        if len(variables) == 0 and unknown_variables in ['ignore', 'skip']:
            return
        pipeline_schedule = self._obj.pipelineschedules.get(pipeline_schedule.id)
        current_variables = dict([[current_variable['key'], current_variable]
                                  for current_variable in pipeline_schedule.attributes['variables']])
        target_variables = dict([[target_variable['key'], deepcopy(target_variable)]
                                 for target_variable in variables])
        # We first check for already existing variables
        for current_variable_key, current_variable in sorted(current_variables.items()):
            if current_variable_key in target_variables:
                for target_variable_param_name, target_variable_param_value in (
                    target_variables[current_variable_key].items()
                ):
                    try:
                        current_variable_param_value = current_variable[target_variable_param_name]
                    except KeyError:
                        logger.info('[%s] NOT Changing pipeline schedule variable %s %s: %s -> %s '
                                    '(current value is not available)',
                                    self._name, current_variable_key, target_variable_param_name,
                                    None, target_variable_param_value)
                        continue
                    if current_variable_param_value != target_variable_param_value:
                        if dry_run:
                            logger.info('[%s] NOT Changing pipeline schedule variable %s %s: %s -> %s (dry-run)',
                                        self._name, current_variable_key, target_variable_param_name,
                                        current_variable_param_value, target_variable_param_value)
                        else:
                            logger.info('[%s] Changing pipeline schedule variable %s %s: %s -> %s',
                                        self._name, current_variable_key, target_variable_param_name,
                                        current_variable_param_value, target_variable_param_value)
                            new_data = {'key': current_variable_key,
                                        'value': target_variables[current_variable_key]['value']}
                            new_data[target_variable_param_name] = target_variable_param_value
                            pipeline_schedule.variables.update(current_variable_key, new_data)
                target_variables.pop(current_variable_key)
            else:
                if unknown_variables in ['delete', 'remove']:
                    if dry_run:
                        logger.info('[%s] NOT Removing pipeline schedule variable %s (dry-run)',
                                    self._name, current_variable_key)
                    else:
                        logger.info('[%s] Removing pipeline schedule variable %s',
                                    self._name, current_variable_key)
                        pipeline_schedule.variables.delete(current_variable_key)
                elif unknown_variables not in ['ignore', 'skip']:
                    logger.warning('[%s] NOT Removing variable: %s',
                                   self._name, current_variable_key)
        # Remaining variables
        for target_variable_name, target_variable in sorted(target_variables.items()):
            if dry_run:
                logger.info('[%s] NOT Adding pipeline schedule variable %s: %s -> %s (dry-run)',
                            self._name, target_variable_name, None, target_variable)
            else:
                logger.info('[%s] Adding pipeline schedule variable %s: %s -> %s',
                            self._name, target_variable_name, None, target_variable)
                pipeline_schedule.variables.create(target_variable)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import inspect
import os
import re

from vcr import VCR


def _gitlabracadabra_func_path_generator(function):
    func_file = inspect.getfile(function)
    method_name = function.__name__  # test_no_create
    instance_name = function.__self__.__class__.__name__  # TestUser
    fixture_name = (re.sub(r'^Test', r'', instance_name) +
                    '_' +
                    re.sub(r'^test_', r'', method_name) +
                    '.yaml')
    return os.path.join(os.path.dirname(func_file),
                        'fixtures',
                        fixture_name,
                        )


def _gitlabracadabra_uri_matcher(r1, r2):
    r1_uri = r1.uri
    r2_uri = r2.uri
    # Workaround 'all=True' in API calls
    # with python-gitlab < 1.8.0
    # See https://github.com/python-gitlab/python-gitlab/pull/701
    if r1_uri.endswith('all=True'):
        r1_uri = r1_uri[0:-9]
    # Ignore host and port
    r1_uri = re.sub('http://[^:/]+(:\\d+)?/', 'http://localhost/', r1_uri)
    r2_uri = re.sub('http://[^:/]+(:\\d+)?/', 'http://localhost/', r2_uri)
    return r1_uri == r2_uri


my_vcr = VCR(
    match_on=['method', 'gitlabracadabra_uri', 'body'],
    func_path_generator=_gitlabracadabra_func_path_generator,
    record_mode='once',  # change to 'new_episodes' to append to existing fixtures
    inject_cassette=True,
)
my_vcr.register_matcher('gitlabracadabra_uri', _gitlabracadabra_uri_matcher)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from copy import deepcopy


logger = logging.getLogger(__name__)


class VariablesMixin(object):
    """Object with variables."""

    """_process_variables()

    Process the variables param.
    """
    def _process_variables(self, param_name, param_value, dry_run=False, skip_save=False):
        assert param_name == 'variables'  # noqa: S101
        assert not skip_save  # noqa: S101
        unknown_variables = self._content.get('unknown_variables', 'warn')
        current_variables = dict([[current_variable.key, current_variable]
                                  for current_variable in self._obj.variables.list(all=True)])
        target_variables = dict([[target_variable['key'], deepcopy(target_variable)]
                                 for target_variable in param_value])
        # We first check for already existing variables
        for current_variable_key, current_variable in sorted(current_variables.items()):
            if current_variable_key in target_variables:
                for target_variable_param_name, target_variable_param_value in (
                    target_variables[current_variable_key].items()
                ):
                    try:
                        current_variable_param_value = getattr(current_variable, target_variable_param_name)
                    except AttributeError:
                        logger.info('[%s] NOT Changing variable %s %s: %s -> %s (current value is not available)',
                                    self._name, current_variable_key, target_variable_param_name,
                                    None, target_variable_param_value)
                        continue
                    if current_variable_param_value != target_variable_param_value:
                        if dry_run:
                            logger.info('[%s] NOT Changing variable %s %s: %s -> %s (dry-run)',
                                        self._name, current_variable_key, target_variable_param_name,
                                        current_variable_param_value, target_variable_param_value)
                        else:
                            logger.info('[%s] Changing variable %s %s: %s -> %s',
                                        self._name, current_variable_key, target_variable_param_name,
                                        current_variable_param_value, target_variable_param_value)
                            setattr(current_variable, target_variable_param_name, target_variable_param_value)
                            current_variable.save()
                target_variables.pop(current_variable_key)
            else:
                if unknown_variables in ['delete', 'remove']:
                    if dry_run:
                        logger.info('[%s] NOT Removing variable %s (dry-run)',
                                    self._name, current_variable_key)
                    else:
                        logger.info('[%s] Removing variable %s',
                                    self._name, current_variable_key)
                        current_variable.delete()
                elif unknown_variables not in ['ignore', 'skip']:
                    logger.warning('[%s] NOT Removing variable: %s',
                                   self._name, current_variable_key)
        # Remaining variables
        for target_variable_name, target_variable in sorted(target_variables.items()):
            if dry_run:
                logger.info('[%s] NOT Adding variable %s: %s -> %s (dry-run)',
                            self._name, target_variable_name, None, target_variable)
            else:
                logger.info('[%s] Adding variable %s: %s -> %s',
                            self._name, target_variable_name, None, target_variable)
                self._obj.variables.create(target_variable)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from time import time
from typing import TYPE_CHECKING
from urllib.request import parse_http_list, parse_keqv_list

from requests import Response, Session, codes
from requests.structures import CaseInsensitiveDict

from gitlabracadabra.auth_info import AuthInfo
from gitlabracadabra.containers.const import DOCKER_HOSTNAME, DOCKER_REGISTRY
from gitlabracadabra.containers.scope import Scope


if TYPE_CHECKING:
    from typing import Any, Dict, List, MutableMapping, Optional, Set, Tuple, Union

    from requests.auth import AuthBase

    Params = Optional[  # noqa: WPS110
        MutableMapping[
            str,
            Union[str, List[str]],
        ]
    ]
    SimpleParams = Dict[str, Union[str, List[str]]]


class Token(object):
    """JWT Token."""

    def __init__(
        self,
        token: str,
        expires_in: int,
    ) -> None:
        """Instanciate a token.

        Args:
            token: Token.
            expires_in: Expires in x seconds.
        """
        minimum_token_lifetime_seconds = 60

        self._token = token
        self._expires_in = expires_in
        if self._expires_in < minimum_token_lifetime_seconds:
            self._expires_in = minimum_token_lifetime_seconds

        # We ignore issued_at property, and use local time instead
        self._issued_at = time()

    @property
    def token(self) -> str:
        """Get token.

        Returns:
            The token.
        """
        return self._token

    @property
    def expiration_time(self) -> float:
        """Get expiration time.

        Returns:
            Expiration time.
        """
        return self._issued_at + self._expires_in

    def is_expired(self) -> bool:
        """Check if token is expired.

        Returns:
            True if token is expired.
        """
        return time() >= self.expiration_time


class RegistrySession(object):  # noqa: WPS214
    """Container registry HTTP methods."""

    def __init__(self, hostname: str, auth_info: Optional[AuthInfo] = None) -> None:
        """Instanciate a registry connection.

        Args:
            hostname: fqdn of a registry.
            auth_info: Authentication information.
        """
        if hostname == DOCKER_HOSTNAME:
            self._hostname = DOCKER_REGISTRY
        else:
            self._hostname = hostname
        self._auth_info = auth_info or AuthInfo()
        self._scheme = 'https'
        self._session = Session()
        self._session.headers = CaseInsensitiveDict({'Docker-Distribution-Api-Version': 'registry/2.0'})
        # Tokens, by set of scopes (as query string or None for all scope)
        self._tokens: Dict[Optional[str], Token] = {}
        # Cache where blobs are present
        # Key is digest, value is a list of manifest names
        # Used in WithBlobs
        self._blobs: Dict[str, List[str]] = {}

    def __del__(self) -> None:  # noqa:WPS603
        """Destroy a registry connection."""
        self._session.close()

    @property
    def hostname(self) -> str:
        """Get hostname.

        Returns:
            The registry hostname.
        """
        if self._hostname == DOCKER_REGISTRY:
            return DOCKER_HOSTNAME
        return self._hostname

    def request(  # noqa:WPS211
        self,
        method: str,
        url: str,
        *,
        scopes: Optional[Set[Scope]] = None,
        params: Params = None,  # noqa: WPS110
        data: Any = None,  # noqa: WPS110
        headers: Optional[Dict[str, str]] = None,
        content_type: Optional[str] = None,
        accept: Optional[Tuple[str, ...]] = None,
        auth: Optional[AuthBase] = None,
        stream: Optional[bool] = None,
        raise_for_status: bool = True,
    ) -> Response:
        """Send an HTTP GET request.

        Args:
            method: HTTP method.
            url: Either a path or a full url.
            scopes: An optional set of scopes.
            params: query string params.
            data: Request body stream.
            headers: Request headers.
            content_type: Uploaded MIME type.
            accept: An optional list of accepted mime-types.
            auth: HTTPBasicAuth.
            stream: Stream the response.
            raise_for_status: Raises `requests.HTTPError`, if one occurred.

        Returns:
            A Response.
        """
        if url.startswith('/'):
            url = '{0}://{1}{2}'.format(self._scheme, self._hostname, url)
        if headers:
            headers = headers.copy()
        else:
            headers = {}
        if accept:
            headers['Accept'] = ', '.join(accept)
        if content_type:
            headers['Content-Type'] = content_type

        self._connect(scopes)
        response = self._session.request(
            method,
            url,
            # https://github.com/python/typeshed/pull/5187
            params=params,  # type: ignore
            data=data,
            headers=headers,
            auth=auth,
            stream=stream,
        )
        if raise_for_status:
            response.raise_for_status()
        return response

    def _connect(self, scopes: Optional[Set[Scope]]) -> None:
        """Connect.

        Args:
            scopes: An optional set of scopes.
        """
        if scopes is None or None in self._tokens:
            self._set_session_auth()
            return
        token = self._get_token(scopes)
        if token:
            self._set_session_auth(token=token)
            return
        response = self.request('get', '/v2/', raise_for_status=False)
        if response.status_code == codes['ok']:
            one_hour = 3600
            self._tokens[None] = Token('no_auth', one_hour)
            return
        if response.status_code == codes['unauthorized']:
            if response.headers['Www-Authenticate'].startswith('Bearer '):
                self._get_bearer_token(response, scopes)
                self._set_session_auth(token=self._get_token(scopes))
                return
        response.raise_for_status()

    def _set_session_auth(self, token: Optional[Token] = None) -> None:
        if token is None:
            if 'Authorization' in self._session.headers:
                self._session.headers.pop('Authorization')
        else:
            self._session.headers['Authorization'] = 'Bearer {0}'.format(token.token)

    def _get_bearer_token(self, response: Response, scopes: Set[Scope]) -> None:
        challenge_parameters = self._get_challenge_parameters(response)
        get_params: SimpleParams = {}
        if 'service' in challenge_parameters:
            get_params['service'] = challenge_parameters.get('service', 'unknown')
        get_params['scope'] = []
        for scope in sorted(scopes):
            get_params['scope'].append(  # type: ignore
                'repository:{0}:{1}'.format(scope.remote_name, scope.actions),
            )
        response = self.request(
            'get',
            challenge_parameters['realm'],
            params=get_params,
            headers=self._auth_info.headers,
            auth=self._auth_info.auth,
        )
        json = response.json()
        self._set_token(
            scopes,
            Token(
                str(json.get('token', json.get('access_token', ''))),
                int(json.get('expires_in', 0)),
            ),
        )

    def _get_challenge_parameters(self, response: Response) -> Dict[str, str]:
        _, _, challenge = response.headers['Www-Authenticate'].partition('Bearer ')
        return parse_keqv_list(parse_http_list(challenge))

    def _get_token(self, scopes: Set[Scope]) -> Optional[Token]:
        key = self._scopes_hash(scopes)
        token = self._tokens.get(key)
        if token and token.is_expired():
            self._tokens.pop(key)
            return None
        return token

    def _set_token(self, scopes: Set[Scope], token: Token) -> None:
        self._tokens[self._scopes_hash(scopes)] = token

    def _scopes_hash(self, scopes: Set[Scope]) -> Optional[str]:
        return ','.join(map(str, sorted(scopes)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from copy import deepcopy


logger = logging.getLogger(__name__)


class BoardsMixin(object):
    """Object with boards."""

    """_process_boards()

    Process the boards param.
    """
    def _process_boards(self, param_name, param_value, dry_run=False, skip_save=False):
        assert param_name == 'boards'  # noqa: S101
        assert not skip_save  # noqa: S101

        unknown_boards = self._content.get('unknown_boards', 'warn')
        unknown_board_lists = self._content.get('unknown_board_lists', 'delete')

        current_boards = dict([[current_board.name, current_board]
                              for current_board in self._obj.boards.list(all=True)])
        target_boards = dict([[target_board['name'], deepcopy(target_board)]
                             for target_board in param_value])
        target_boards_by_old_name = dict([[target_board['old_name'], deepcopy(target_board)]
                                         for target_board in param_value if 'old_name' in target_board])

        # We do not reuse labels from LabelsMixin, to have fresh data
        try:
            self._current_labels = dict([[current_label.name, current_label]
                                        for current_label in self._obj.labels.list(all=True)])
        except AttributeError:
            # https://github.com/python-gitlab/python-gitlab/pull/847
            logger.error('[%s] Unable to manage labels: %s',
                         self._name, 'group labels requires python-gitlab >= 1.11.0')
            return

        # We first check for already existing boards
        for current_board_name, current_board in sorted(current_boards.items()):
            if current_board_name in target_boards:
                target_board = target_boards.pop(current_board_name)
            elif current_board_name in target_boards_by_old_name:
                target_board = target_boards_by_old_name[current_board_name]
                target_boards.pop(target_board['name'])
                if dry_run:
                    logger.info('[%s] NOT Changing board %s %s: %s -> %s (dry-run)',
                                self._name, current_board_name,
                                'name', current_board_name,
                                target_board['name'])
                else:
                    logger.info('[%s] Changing board %s %s: %s -> %s',
                                self._name, current_board_name,
                                'name', current_board_name,
                                target_board['name'])
                    current_board.name = target_board['name']
                    try:
                        current_board.save()
                    except KeyError:
                        # v1.10.0 - feat: add support for board update
                        # https://github.com/python-gitlab/python-gitlab/pull/804
                        logger.error('[%s] Unable to update board: %s',
                                     self._name, 'board update requires python-gitlab >= 1.10.0')
                        return
            else:
                target_board = None
            if target_board:
                for target_board_param_name, target_board_param_value in target_board.items():
                    if target_board_param_name == 'lists':
                        unknown_lists = target_board.get('unknown_lists', unknown_board_lists)
                        self._handle_board_lists(current_board, target_board_param_value, unknown_lists, dry_run)
                        continue
                    if target_board_param_name in ['old_name', 'unknown_lists']:
                        continue
                    try:
                        current_board_param_value = getattr(current_board, target_board_param_name)
                    except AttributeError:
                        logger.info('[%s] NOT Changing boards %s %s: %s -> %s '
                                    '(current value is not available)',
                                    self._name, current_board_name,
                                    target_board_param_name, None, target_board_param_value)
                        continue
                    if current_board_param_value != target_board_param_value:
                        if dry_run:
                            logger.info('[%s] NOT Changing board %s %s: %s -> %s (dry-run)',
                                        self._name, current_board_name,
                                        target_board_param_name, current_board_param_value,
                                        target_board_param_value)
                        else:
                            logger.info('[%s] Changing board %s %s: %s -> %s',
                                        self._name, current_board_name,
                                        target_board_param_name, current_board_param_value,
                                        target_board_param_value)
                            setattr(current_board, target_board_param_name,
                                    target_board_param_value)
                            try:
                                current_board.save()
                            except KeyError:
                                # v1.10.0 - feat: add support for board update
                                # https://github.com/python-gitlab/python-gitlab/pull/804
                                logger.error('[%s] Unable to update board: %s',
                                             self._name, 'board update requires python-gitlab >= 1.10.0')
                                return
            else:
                if unknown_boards in ['delete', 'remove']:
                    if dry_run:
                        logger.info('[%s] NOT Removing board %s (dry-run)',
                                    self._name, current_board_name)
                    else:
                        logger.info('[%s] Removing board %s',
                                    self._name, current_board_name)
                        current_board.delete()
                elif unknown_boards not in ['ignore', 'skip']:
                    logger.warning('[%s] NOT Removing board: %s',
                                   self._name, current_board_name)
        # Remaining boards
        for target_board_name, target_board in sorted(target_boards.items()):
            target_board_lists = target_board.pop('lists', [])
            unknown_lists = target_board.pop('unknown_lists', unknown_board_lists)
            if dry_run:
                logger.info('[%s] NOT Adding board %s: %s -> %s (dry-run)',
                            self._name, target_board_name, None, target_board)
            else:
                logger.info('[%s] Adding board %s: %s -> %s',
                            self._name, target_board_name, None, target_board)
                board = self._obj.boards.create(target_board)
                self._handle_board_lists(board, target_board_lists, unknown_lists, dry_run)

    """_handle_board_lists()

    Handle board lists.
    """
    def _handle_board_lists(self, board, target_lists, unknown_lists, dry_run=False):
        if len(target_lists) == 0 and unknown_lists in ['ignore', 'skip']:
            return
        board = self._obj.boards.get(board.id)
        current_lists = dict([[current_list.label['name'], current_list]
                              for current_list in board.lists.list()])
        # We first add any missing list
        for target_list_position, target_list in enumerate(target_lists):
            if target_list['label'] in current_lists:
                current_list = current_lists[target_list['label']]
                if target_list_position != current_list.position:
                    old_list_position = current_list.position
                    # Boards::Lists::MoveService
                    # https://gitlab.com/gitlab-org/gitlab/-/blob/master/app/services/boards/lists/move_service.rb#L29
                    if old_list_position < target_list_position:
                        self._decrement_intermediate_lists(current_lists, old_list_position, target_list_position)
                    else:
                        self._increment_intermediate_lists(current_lists, old_list_position, target_list_position)
                    current_list.position = target_list_position
                    if dry_run:
                        logger.info('[%s] NOT Changing board list position %s: %s -> %s (dry-run)',
                                    self._name, target_list,
                                    current_list.position, target_list_position)
                    else:
                        logger.info('[%s] Changing board list position %s: %s -> %s (dry-run)',
                                    self._name, target_list,
                                    current_list.position, target_list_position)
                        current_list.save()
            else:
                if target_list['label'] not in self._current_labels:
                    logger.warning('[%s] Label not found for board %s: %s',
                                   self._name, board.name, target_list['label'])
                    continue
                label_id = self._current_labels[target_list['label']].id
                if dry_run:
                    logger.info('[%s] NOT Adding board list %s (dry-run)',
                                self._name, target_list)
                else:
                    logger.info('[%s] Adding board list %s',
                                self._name, target_list)
                    board.lists.create({'label_id': label_id})

        if unknown_lists in ['ignore', 'skip']:
            return
        # Delete any remaining lists
        for current_list_label, current_list in sorted(current_lists.items()):
            current_list_dict = {'label': current_list_label}
            if current_list_dict in target_lists:
                continue

            if unknown_lists in ['delete', 'remove']:
                if dry_run:
                    logger.info('[%s] NOT Removing board list %s (dry-run)',
                                self._name, current_list_dict)
                else:
                    logger.info('[%s] Removing board list %s',
                                self._name, current_list_dict)
                    current_list.delete()
            elif unknown_lists not in ['ignore', 'skip']:
                logger.warning('[%s] NOT Removing board list: %s',
                               self._name, current_list_dict)

    """_decrement_intermediate_lists()

    Decrement intermediate lists.
    """
    def _decrement_intermediate_lists(self, lists, old_position, new_position):
        for _list_name, list_object in lists.items():
            if list_object.position > old_position and list_object.position <= new_position:
                list_object.position = list_object.position - 1

    """_increment_intermediate_lists()

    Increment intermediate lists.
    """
    def _increment_intermediate_lists(self, lists, old_position, new_position):
        for _list_name, list_object in lists.items():
            if list_object.position >= new_position and list_object.position < old_position:
                list_object.position = list_object.position + 1



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import annotations

from typing import TYPE_CHECKING

from gitlab import Gitlab

from requests.auth import HTTPBasicAuth

from gitlabracadabra.auth_info import AuthInfo
from gitlabracadabra.gitlab.group_cache import GroupCache
from gitlabracadabra.gitlab.user_cache import UserCache


if TYPE_CHECKING:
    from typing import List, Optional


class PyGitlab(object):
    """Python-Gitlab wrapper."""

    def __init__(
        self,
        gitlab_id: Optional[str],
        config_files: Optional[List[str]],
        debug: bool,
        auth: bool = True,
    ) -> None:
        """Initialize a Python-Gitlab wrapper.

        Args:
            gitlab_id: Section in python-gitlab config files.
            config_files: None or list of configuration files.
            debug: True to enable debugging.
            auth: True to authenticate on creation.
        """
        self._gitlab_id = gitlab_id
        self._config_files = config_files
        self._debug = debug
        self._gl = Gitlab.from_config(self.gitlab_id, self._config_files)
        if auth:
            if self.pygitlab.private_token or self.pygitlab.oauth_token:
                self.pygitlab.auth()
        if self._debug:
            self.pygitlab.enable_debug()

        self.group_cache = GroupCache(self)
        self.user_cache = UserCache(self)

    @property
    def gitlab_id(self) -> Optional[str]:
        """Get Gitlab id (section in python-gitlab config files).

        Returns:
            A string.
        """
        return self._gitlab_id

    @property
    def pygitlab(self) -> Gitlab:
        """Get python-gitlab object.

        Returns:
            A gitlab.Gitlab object.
        """
        return self._gl

    @property
    def registry_auth_info(self) -> AuthInfo:
        """Get Registry Authentication information.

        Returns:
            A dict, with 'headers' and 'auth' to pass to requests.

        Raises:
            ValueError: No auth info.
        """
        if self.pygitlab.private_token:
            return AuthInfo(auth=HTTPBasicAuth('personal-access-token', self.pygitlab.private_token))
        if self.pygitlab.oauth_token:
            return AuthInfo(auth=HTTPBasicAuth('oauth2', self.pygitlab.oauth_token))
        if self.pygitlab.job_token:
            return AuthInfo(auth=HTTPBasicAuth('gitlab-ci-token', self.pygitlab.job_token))
        if self.pygitlab.http_username:
            return AuthInfo(auth=HTTPBasicAuth(self.pygitlab.http_username, self.pygitlab.http_password))
        raise ValueError('No auth info')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from contextlib import contextmanager
from os.path import isfile
from typing import TYPE_CHECKING, Generator
from unittest.mock import call, patch

from gitlabracadabra.containers.blob import Blob
from gitlabracadabra.containers.const import DOCKER_HOSTNAME, DOCKER_MANIFEST_SCHEMA2, DOCKER_MANIFEST_SCHEMA2_LIST
from gitlabracadabra.containers.manifest import Manifest
from gitlabracadabra.containers.registry import Registry
from gitlabracadabra.containers.with_digest import WithDigest
from gitlabracadabra.tests import my_vcr
from gitlabracadabra.tests.case import TestCase


if TYPE_CHECKING:
    from logging import Logger


VERIFY_DIGEST_METHOD = '_verify_digest'
REGISTRY_IMPORTER_LOGGER = 'gitlabracadabra.containers.registry_importer.logger'
REGISTRY_IMPORTER_MESSAGE = '%s%s %simported as %s:%s (%s, %s, %s)'
BLOBS2_0_0_MESSAGE = '2+0+0=2 uploaded+mounted+existing blobs'

DEBIAN_IMAGE = 'library/debian'
BUSYBOX_IMAGE = 'library/busybox'
BUSYBOX_IMAGE_FULL = 'docker.io/library/busybox:latest@{0}'
LATEST = 'latest'


class TestRegistry(TestCase):  # noqa: WPS214
    def destination_registry(self):
        registry = Registry('localhost:5000')
        registry._scheme = 'http'  # noqa: WPS437
        return registry

    def assert_with_digest(self, registry, cacheable, digest, size, mime_type):  # noqa:WPS211
        self.assertIsInstance(cacheable, WithDigest)
        self.assertEqual(cacheable.registry, registry)
        self.assertEqual(cacheable.digest, digest)
        self.assertEqual(cacheable.size, size)
        self.assertEqual(cacheable.mime_type, mime_type)

    def assert_manifest(self, registry, manifest, digest, size, mime_type):  # noqa:WPS211
        self.assert_with_digest(registry, manifest, digest, size, mime_type)
        self.assertIsInstance(manifest, Manifest)
        self.assertIsInstance(manifest.json, dict)
        self.assertTrue(isfile(manifest.cache_path))
        self.assertEqual(manifest.json['mediaType'], mime_type)
        self.assertEqual(manifest.json['schemaVersion'], 2)

    @contextmanager
    def mocked_objects(self) -> Generator[Logger, None, None]:
        with patch.object(WithDigest, VERIFY_DIGEST_METHOD) as verify_digest_mock:
            verify_digest_mock.return_value = None
            with patch(REGISTRY_IMPORTER_LOGGER, autospec=True) as logger:
                yield logger

    @my_vcr.use_cassette
    def test_manifest(self, cass):
        digest = 'sha256:102ab2db1ad671545c0ace25463c4e3c45f9b15e319d3a00a1b2b085293c27fb'
        registry = Registry(DOCKER_HOSTNAME)  # noqa: WPS204
        manifest = registry.manifest(DEBIAN_IMAGE, digest)
        self.assert_manifest(
            registry,
            manifest,
            digest,
            529,  # noqa:WPS432
            DOCKER_MANIFEST_SCHEMA2,
        )
        self.assertIsInstance(manifest.json['config'], dict)
        self.assertIsInstance(manifest.json['layers'], list)
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_manifest_list(self, cass):
        registry = Registry(DOCKER_HOSTNAME)
        manifest_list = registry.manifest(DEBIAN_IMAGE)
        self.assert_manifest(
            registry,
            manifest_list,
            'sha256:1092695e843ad975267131f27a2b523128c4e03d2d96574bbdd7cf949ed51475',
            1854,  # noqa:WPS432
            DOCKER_MANIFEST_SCHEMA2_LIST,
        )
        self.assertIsInstance(manifest_list.json['manifests'], list)

        manifests = manifest_list.manifests()
        self.assertIsInstance(manifests, list)
        first = manifests[0]
        self.assertFalse(isfile(first.cache_path))
        self.assert_manifest(
            registry,
            first,
            'sha256:102ab2db1ad671545c0ace25463c4e3c45f9b15e319d3a00a1b2b085293c27fb',
            529,  # noqa:WPS432
            DOCKER_MANIFEST_SCHEMA2,
        )
        self.assertIsInstance(first, Manifest)
        self.assertEqual(first.platform, {'architecture': 'amd64', 'os': 'linux'})
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_blob(self, cass):
        manifest_name = DEBIAN_IMAGE
        digest = 'sha256:5890f8ba95f680c87fcf89e51190098641b4f646102ce7ca906e7f83c84874dc'
        registry = Registry(DOCKER_HOSTNAME)
        blob = registry.blob(manifest_name, digest)
        self.assertIsInstance(blob, Blob)
        self.assertEqual(blob.registry, registry)
        self.assertEqual(blob.manifest_name, manifest_name)
        self.assertEqual(blob.digest, digest)
        self.assertFalse(isfile(blob.cache_path))
        self.assertEqual(blob.size, 1459)  # noqa:WPS432
        self.assertFalse(isfile(blob.cache_path))
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_import_manifest(self, cass):
        digest = 'sha256:74e4a68dfba6f40b01787a3876cc1be0fb1d9025c3567cf8367c659f2187234f'
        registry = Registry(DOCKER_HOSTNAME)
        registry2 = self.destination_registry()
        manifest = Manifest(
            registry,
            BUSYBOX_IMAGE,
            digest,
        )
        with self.mocked_objects() as logger:
            registry2.import_manifest(manifest)
            logger.info.assert_has_calls([call(
                REGISTRY_IMPORTER_MESSAGE,
                '',
                BUSYBOX_IMAGE_FULL.format(digest),
                '',
                BUSYBOX_IMAGE,
                LATEST,
                '1+0=1 uploaded+existing manifests',
                BLOBS2_0_0_MESSAGE,
                '766112+0+0=766112 uploaded+mounted+existing blobs size',
            )])
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_import_manifest_already_exists(self, cass):
        digest = 'sha256:74e4a68dfba6f40b01787a3876cc1be0fb1d9025c3567cf8367c659f2187234f'
        registry = Registry(DOCKER_HOSTNAME)
        registry2 = self.destination_registry()
        manifest = Manifest(
            registry,
            BUSYBOX_IMAGE,
            digest,
        )
        with self.mocked_objects() as logger:
            registry2.import_manifest(manifest)
            logger.info.assert_has_calls([])
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_import_manifest_list(self, cass):
        digest = 'sha256:c6b45a95f932202dbb27c31333c4789f45184a744060f6e569cc9d2bf1b9ad6f'
        registry = Registry(DOCKER_HOSTNAME)
        registry2 = self.destination_registry()
        manifest = Manifest(
            registry,
            BUSYBOX_IMAGE,
            tag=LATEST,
        )
        with self.mocked_objects() as logger:
            registry2.import_manifest(manifest)
            logger.info.assert_has_calls([call(
                REGISTRY_IMPORTER_MESSAGE,
                '',
                BUSYBOX_IMAGE_FULL.format(digest),
                '',
                BUSYBOX_IMAGE,
                LATEST,
                '1+0=1 uploaded+existing manifests',
                BLOBS2_0_0_MESSAGE,
                '766112+0+0=766112 uploaded+mounted+existing blobs size',
            )])
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_import_manifest_list_dry_run(self, cass):
        digest = 'sha256:c6b45a95f932202dbb27c31333c4789f45184a744060f6e569cc9d2bf1b9ad6f'
        registry = Registry(DOCKER_HOSTNAME)
        registry2 = self.destination_registry()
        manifest = Manifest(
            registry,
            BUSYBOX_IMAGE,
            tag=LATEST,
        )
        with self.mocked_objects() as logger:
            registry2.import_manifest(manifest, dry_run=True)
            logger.info.assert_has_calls([call(
                REGISTRY_IMPORTER_MESSAGE,
                '',
                BUSYBOX_IMAGE_FULL.format(digest),
                'NOT ',
                BUSYBOX_IMAGE,
                LATEST,
                '1+0=1 uploaded+existing manifests',
                BLOBS2_0_0_MESSAGE,
                '766112+0+0=766112 uploaded+mounted+existing blobs size',
            )])
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_import_manifest_list_all(self, cass):
        digest = 'sha256:c6b45a95f932202dbb27c31333c4789f45184a744060f6e569cc9d2bf1b9ad6f'
        registry = Registry(DOCKER_HOSTNAME)
        registry2 = self.destination_registry()
        manifest = Manifest(
            registry,
            BUSYBOX_IMAGE,
            tag=LATEST,
        )
        with self.mocked_objects() as logger:
            registry2.import_manifest(manifest, 'newname', 'newtag', platform='all')
            logger.info.assert_has_calls([call(
                REGISTRY_IMPORTER_MESSAGE,
                '',
                BUSYBOX_IMAGE_FULL.format(digest),
                '',
                'newname',
                'newtag',
                '2+8=10 uploaded+existing manifests',
                BLOBS2_0_0_MESSAGE,
                '2140766+0+0=2140766 uploaded+mounted+existing blobs size',
            )])
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_import_manifest_not_found(self, cass):
        registry = Registry(DOCKER_HOSTNAME)
        registry2 = self.destination_registry()
        manifest = Manifest(
            registry,
            BUSYBOX_IMAGE,
            tag='not_found',
        )
        url = 'https://registry-1.docker.io/v2/library/busybox/manifests/not_found'
        with self.mocked_objects() as logger:
            registry2.import_manifest(manifest)
            logger.warning.assert_has_calls([call(
                '%s%s NOT imported as %s:%s: %s',
                '',
                'docker.io/library/busybox:not_found',
                'library/busybox',
                'not_found',
                '404 Client Error: Not Found for url: {0}'.format(url),
            )])
        self.assertTrue(cass.all_played)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import namedtuple

from gitlabracadabra.containers.const import DOCKER_HOSTNAME, DOCKER_REGISTRY
from gitlabracadabra.containers.registries import Registries
from gitlabracadabra.tests.case import TestCase


TestData = namedtuple(
    'TestData',
    ['input_name', 'hostname', 'manifest_name', 'tag', 'reference_tag', 'digest', 'full_reference', 'short_reference'],
)

LATEST = 'latest'
TEST_DATA = (
    TestData(
        'my/repo',
        DOCKER_HOSTNAME,
        'my/repo',
        LATEST,
        None,
        None,
        'docker.io/my/repo',
        'my/repo',
    ),
    TestData(
        'debian',
        DOCKER_HOSTNAME,
        'library/debian',
        LATEST,
        None,
        None,
        'docker.io/library/debian',
        'debian',
    ),
    TestData(
        'localhost/foo/bar:v1.0',
        'localhost',
        'foo/bar',
        'v1.0',
        'v1.0',
        None,
        'localhost/foo/bar:v1.0',
        'localhost/foo/bar:v1.0',
    ),
    TestData(
        'example.org/foo/baz:v2.3.4',
        'example.org',
        'foo/baz',
        'v2.3.4',
        'v2.3.4',
        None,
        'example.org/foo/baz:v2.3.4',
        'example.org/foo/baz:v2.3.4',
    ),
    TestData(
        'registry:5000/foo/bar:latest',
        'registry:5000',
        'foo/bar',
        LATEST,
        LATEST,
        None,
        'registry:5000/foo/bar:latest',
        'registry:5000/foo/bar',
    ),
    TestData(
        'busybox:latest@sha256:74e4a68dfba6f40b01787a3876cc1be0fb1d9025c3567cf8367c659f2187234f',
        DOCKER_HOSTNAME,
        'library/busybox',
        LATEST,
        LATEST,
        'sha256:74e4a68dfba6f40b01787a3876cc1be0fb1d9025c3567cf8367c659f2187234f',
        'docker.io/library/busybox:latest@sha256:74e4a68dfba6f40b01787a3876cc1be0fb1d9025c3567cf8367c659f2187234f',
        'busybox@sha256:74e4a68dfba6f40b01787a3876cc1be0fb1d9025c3567cf8367c659f2187234f',
    ),
)


class TestRegistries(TestCase):
    def test_singleton(self) -> None:
        registries1 = Registries()
        registries2 = Registries()
        self.assertEqual(id(registries1), id(registries2))

    def test_get_registry(self) -> None:
        registries = Registries()
        docker_registry = registries.get_registry(DOCKER_REGISTRY)
        self.assertEqual(docker_registry.hostname, DOCKER_HOSTNAME)
        docker_registry2 = registries.get_registry(DOCKER_HOSTNAME)
        self.assertEqual(docker_registry2.hostname, DOCKER_HOSTNAME)
        docker_registry3 = registries.get_registry('example.org')
        self.assertEqual(docker_registry3.hostname, 'example.org')

    def test_get_manifest(self) -> None:
        registries = Registries()
        for test_data in TEST_DATA:
            with self.subTest(input_name=test_data.input_name):
                manifest = registries.get_manifest(test_data.input_name)
                self.assertEqual(manifest.registry.hostname, test_data.hostname)
                self.assertEqual(manifest.manifest_name, test_data.manifest_name)
                self.assertEqual(manifest.tag, test_data.tag)
                self.assertEqual(manifest._digest, test_data.digest)  # noqa: WPS437

    def test_short_reference(self) -> None:
        for test_data in TEST_DATA:
            with self.subTest(input_name=test_data.input_name):
                reference = Registries.short_reference(test_data.input_name)
                self.assertEqual(reference, test_data.short_reference)

    def test_full_reference(self) -> None:
        for test_data in TEST_DATA:
            with self.subTest(input_name=test_data.input_name):
                reference = Registries.full_reference(test_data.input_name)
                self.assertEqual(reference, test_data.full_reference)

    def test_full_reference_parts(self) -> None:
        for test_data in TEST_DATA:
            with self.subTest(input_name=test_data.input_name):
                manifest = Registries.full_reference_parts(test_data.input_name)
                self.assertEqual(manifest.hostname, test_data.hostname)
                self.assertEqual(manifest.manifest_name, test_data.manifest_name)
                self.assertEqual(manifest.tag, test_data.reference_tag)
                self.assertEqual(manifest.digest, test_data.digest)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['my_vcr', 'patch', 'patch_open']

from gitlabracadabra.tests.patchfuncs import patch, patch_open
from gitlabracadabra.tests.vcrfuncs import my_vcr



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gitlabracadabra.objects.user import GitLabracadabraUser
from gitlabracadabra.tests import my_vcr
from gitlabracadabra.tests.case import TestCaseWithManager


class TestUser(TestCaseWithManager):
    @my_vcr.use_cassette
    def test_no_create(self, cass):
        obj = GitLabracadabraUser('memory', 'no_create_user', {})
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_create(self, cass):
        obj = GitLabracadabraUser('memory', 'create_user', {
            'create_object': True,
            'email': 'create_user@example.org',
            'name': 'Create User',
            'password': 'P@ssw0rdNot24get',
        })
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_delete(self, cass):
        obj = GitLabracadabraUser('memory', 'delete_this_user', {'delete_object': True})
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_exists(self, cass):
        obj = GitLabracadabraUser('memory', 'user_exists', {})
        obj.process()
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_simple_parameters(self, cass):
        obj = GitLabracadabraUser('memory', 'user_simple_parameters', {
            'name': 'user-with-simple-parameters',
            'email': 'user-with-simple-parameters@example.org',
            'skip_confirmation': True,
            'skip_reconfirmation': True,
            'public_email': 'contact@example.org',
            'password': 'P@ss12345678',
            # 'reset_password'
            'projects_limit': 42,
            'can_create_group': False,
            'admin': True,
            'external': True,
            'shared_runners_minutes_limit': 42,
            'extra_shared_runners_minutes_limit': 42,
            # 'avatar'
            'skype': '12345',
            'linkedin': 'linked_in',
            'twitter': 't_w_i_t_t_e_r',
            'website_url': 'https://example.org',
            'location': 'Nowhere',
            'organization': 'My Corp',
            'bio': 'Not much',
            'private_profile': True,
            'note': 'Fake account?',
            'extern_uid': '12345678',
            'provider': 'github',
            # 'color_scheme_id'
            # 'theme_id'
            # 'force_random_password'
            # 'group_id_for_saml'
            # 'view_diffs_file_by_file'
        })
        self.assertEqual(obj.errors(), [])
        obj.process()
        self.assertTrue(cass.all_played)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from contextlib import contextmanager
from typing import Generator
from unittest.mock import call, patch

from gitlab import Gitlab, __version__ as gitlab_version

from gitlabracadabra.gitlab.connections import GitlabConnections
from gitlabracadabra.tests.case import TestCase
from gitlabracadabra.tests.patchfuncs import patch_open


class TestGitlabConnections(TestCase):
    """Test GitlabConnections class."""

    @contextmanager
    def fake_config(self, call_count: int = 1) -> Generator[GitlabConnections, None, None]:
        """Fake configuration.

        Args:
            call_count: Number of Gitlab objects created.

        Yields:
            The Gitlab connections singleton.
        """
        config_path = '/path/to/python-gitlab.cfg'
        singleton = GitlabConnections()
        singleton.load(None, [config_path], debug=False)
        config = """
            [global]
            default = gitlab

            [gitlab]
            url = https://gitlab.com
            private_token = T0k3n

            [internal]
            url = https://gitlab.example.com
            private_token = n3k0T
        """
        with patch('os.path.exists') as os_path_exists_mock:
            with patch_open({config_path: config}):
                with patch.object(Gitlab, 'auth') as auth_mock:
                    os_path_exists_mock.return_value = True
                    yield singleton
                    if gitlab_version != '1.6.0':
                        self.assertEqual(  # noqa: WPS220
                            os_path_exists_mock.mock_calls,
                            [call(config_path) for _ in range(call_count)],
                        )
                    self.assertEqual(auth_mock.mock_calls, [call() for _ in range(call_count)])

    def test_singleton(self) -> None:
        """Ensure singleton pattern."""
        singleton1 = GitlabConnections()
        singleton2 = GitlabConnections()
        self.assertEqual(id(singleton1), id(singleton2))

    def test_get_connection_none(self) -> None:
        """Get default Gitlab connection."""
        with self.fake_config() as singleton:
            gl1 = singleton.get_connection()
            self.assertEqual(gl1.pygitlab.api_url, 'https://gitlab.com/api/v4')
            self.assertEqual(gl1.pygitlab.private_token, 'T0k3n')
            gl2 = singleton.get_connection(None)
            self.assertEqual(id(gl1), id(gl2))

    def test_get_connection_internal(self) -> None:
        """Get another Gitlab connection."""
        with self.fake_config() as singleton:
            gl1 = singleton.get_connection('internal')
            self.assertEqual(gl1.pygitlab.api_url, 'https://gitlab.example.com/api/v4')
            self.assertEqual(gl1.pygitlab.private_token, 'n3k0T')
            gl2 = singleton.get_connection('internal')
            self.assertEqual(id(gl1), id(gl2))

    def test_get_connection_both(self) -> None:
        """Get several Gitlab connections."""
        with self.fake_config(2) as singleton:
            gl1 = singleton.get_connection()
            gl2 = singleton.get_connection('internal')
            self.assertEqual(gl1.pygitlab.api_url, 'https://gitlab.com/api/v4')
            self.assertEqual(gl1.pygitlab.private_token, 'T0k3n')
            self.assertEqual(gl2.pygitlab.api_url, 'https://gitlab.example.com/api/v4')
            self.assertEqual(gl2.pygitlab.private_token, 'n3k0T')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from requests import HTTPError, codes

from gitlabracadabra.containers.const import DOCKER_MANIFEST_SCHEMA1_SIGNED, DOCKER_MANIFEST_SCHEMA2
from gitlabracadabra.containers.manifest import Manifest
from gitlabracadabra.containers.registry import Registry
from gitlabracadabra.tests import my_vcr
from gitlabracadabra.tests.case import TestCase


class TestManifest(TestCase):
    """Test Manifest class."""

    @my_vcr.use_cassette
    def test_tag_list(self, cass):
        """Test tag_list method.

        Args:
            cass: VCR cassette.
        """
        registry = Registry('docker.io')
        manifest = Manifest(registry, 'library/debian')
        self.assertIsInstance(manifest, Manifest)
        tag_list = manifest.tag_list()
        self.assertIsInstance(tag_list, list)
        self.assertIn('buster', tag_list)
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_not_found(self, cass):
        """Test proper 404 handling.

        Args:
            cass: VCR cassette.
        """
        registry = Registry('docker.io')
        for attr in 'digest', 'size', 'mime_type':
            with self.subTest(attr=attr):
                with self.assertRaises(HTTPError) as cm:
                    manifest = Manifest(registry, 'library/debian', tag='not_found')
                    getattr(manifest, attr)
                self.assertEqual(cm.exception.response.status_code, codes['not_found'])  # noqa: WPS441
        self.assertFalse(Manifest(registry, 'library/debian', tag='not_found').exists())
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_blobs(self, cass):
        """Test blobs method.

        Args:
            cass: VCR cassette.
        """
        registry = Registry('registry.developers.crunchydata.com')
        manifest = Manifest(registry, 'crunchydata/pgo-apiserver', None, tag='centos8-4.6.2')
        self.assertEqual(manifest.mime_type, DOCKER_MANIFEST_SCHEMA2)
        blobs = manifest.blobs()
        self.assertEqual(len(blobs), 17)  # noqa: WPS432
        blob0 = blobs[0]
        self.assertEqual(blob0.manifest_name, manifest.manifest_name)
        self.assertEqual(blob0.mime_type, 'application/vnd.docker.image.rootfs.diff.tar.gzip')
        self.assertEqual(blob0.size, 75181999)  # noqa: WPS432
        self.assertEqual(blob0.digest, 'sha256:7a0437f04f83f084b7ed68ad9c4a4947e12fc4e1b006b38129bac89114ec3621')
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette  # noqa: WPS213
    def test_blobs_manifest_v1(self, cass):  # noqa: WPS213
        """Test blobs method with a manifest v1.

        Args:
            cass: VCR cassette.
        """
        registry = Registry('quay.io')
        manifest = Manifest(registry, 'jetstack/cert-manager-controller', None, tag='v0.1.0')
        self.assertEqual(manifest.mime_type, DOCKER_MANIFEST_SCHEMA1_SIGNED)
        blobs = manifest.blobs()
        self.assertEqual(len(blobs), 7)  # noqa: WPS432
        self.assertEqual(blobs[0].manifest_name, manifest.manifest_name)
        self.assertEqual(blobs[0].mime_type, 'application/octet-stream')
        self.assertEqual(cass.play_count, 3)
        self.assertEqual(blobs[0].size, 32)  # noqa: WPS432
        self.assertEqual(cass.play_count, 4)
        self.assertEqual(blobs[0].digest, 'sha256:a3ed95caeb02ffe68cdd9fd84406680ae93d633cb16422d00e8a7c22955b46d4')
        self.assertTrue(cass.all_played)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import logging
import subprocess  # noqa: S404
from contextlib import suppress
from os.path import isdir
from typing import TYPE_CHECKING, Any, Dict, List, Optional
from urllib.parse import quote, urlparse
from urllib.request import getproxies, proxy_bypass

from pygit2 import GIT_FETCH_PRUNE, LIBGIT2_VER, GitError, RemoteCallbacks, Repository, init_repository

from gitlabracadabra.disk_cache import cache_dir
from gitlabracadabra.gitlab.connections import GitlabConnections
from gitlabracadabra.matchers import Matcher


if TYPE_CHECKING:
    from pygit2 import Reference, UserPass


logger = logging.getLogger(__name__)


class MirrorsMixin(object):
    """Object with mirrors."""

    def _init_repo(self) -> None:
        """Init the cache repository."""
        web_url_slug = quote(self.web_url(), safe='')  # type: ignore
        repo_dir = str(cache_dir('') / web_url_slug)
        if isdir(repo_dir):
            self._repo = Repository(repo_dir)
        else:
            logger.debug(
                '[%s] Creating cache repository in %s',
                self._name,  # type: ignore
                repo_dir,
            )
            self._repo = init_repository(repo_dir, bare=True)
        try:
            self._repo.remotes['gitlab']  # noqa: WPS428
        except KeyError:
            self._repo.remotes.create(
                'gitlab',
                self.web_url(),  # type: ignore
                '+refs/heads/*:refs/remotes/gitlab/heads/*',
            )
            self._repo.remotes.add_fetch('gitlab', '+refs/tags/*:refs/remotes/gitlab/tags/*')
            self._repo.remotes.add_push('gitlab', '+refs/heads/*:refs/heads/*')
            self._repo.remotes.add_push('gitlab', '+refs/tags/*:refs/tags/*')
            self._repo.config['remote.gitlab.mirror'] = True

    def _fetch_remote(self, name: str, credentials: Optional[UserPass] = None) -> None:  # noqa: WPS210,WPS231
        """Fetch the repo with the given name.

        Args:
            name: Remote name.
            credentials: Credentials as pygit2.UserPass.
        """
        url = self._repo.config['remote.{name}.url'.format(name=name)]
        libgit2_workaround = False
        if url.startswith('https://') and LIBGIT2_VER < (0, 28, 0):
            try:
                http_proxy = self._repo.config['http.proxy']
            except KeyError:
                http_proxy = None
            with suppress(KeyError):
                http_proxy = self._repo.config['remote.{name}.proxy'.format(name=name)]
            if http_proxy is None:  # '' being explicitly disabled
                proxies = getproxies()
                parsed = urlparse(url)
                http_proxy = proxies.get(parsed.scheme) or proxies.get('any')
                if proxy_bypass(parsed.hostname):
                    http_proxy = None
            if http_proxy:
                libgit2_workaround = True
        if libgit2_workaround:
            # libgit2 >= 0.28 required for proper HTTP proxy support
            # https://github.com/libgit2/libgit2/pulls/4870
            # https://github.com/libgit2/libgit2/pulls/5052
            logger.warning(
                '[%s] Using git command to fetch remote %s',
                self._name,  # type: ignore
                name,
            )
            subprocess.run(  # noqa: S603,S607
                ['git', 'fetch', '--quiet', '--prune', name],
                cwd=self._repo.path,
                check=True,
            )
        else:
            if credentials:
                pygit2_callbacks = RemoteCallbacks(credentials=credentials)
            else:
                pygit2_callbacks = None
            try:
                # https://gitlab.com/gitlabracadabra/gitlabracadabra/-/issues/25
                self._repo.remotes[name].fetch(
                    refspecs=self._repo.remotes[name].fetch_refspecs,
                    callbacks=pygit2_callbacks,
                    prune=GIT_FETCH_PRUNE,
                    proxy=True,
                )
            except TypeError:
                self._repo.remotes[name].fetch(
                    refspecs=self._repo.remotes[name].fetch_refspecs,
                    callbacks=pygit2_callbacks,
                    prune=GIT_FETCH_PRUNE,
                )

    def _push_remote(self, name: str, refs: List[str], credentials: Optional[UserPass] = None) -> None:
        """Push to the repo with the given name.

        Args:
            name: Remote name.
            refs: refs list.
            credentials: Credentials as pygit2.UserPass.
        """
        if credentials:
            pygit2_callbacks = RemoteCallbacks(credentials=credentials)
        else:
            pygit2_callbacks = None
        try:
            try:  # noqa: WPS505
                # https://gitlab.com/gitlabracadabra/gitlabracadabra/-/issues/25
                self._repo.remotes[name].push(refs, callbacks=pygit2_callbacks, proxy=True)
            except TypeError:
                self._repo.remotes[name].push(refs, callbacks=pygit2_callbacks)
        except GitError as err:
            logger.error(  # noqa: G200
                '[%s] Unable to push remote=%s refs=%s: %s',
                self._name,  # type: ignore
                name,
                ','.join(refs),
                err,
            )

    def _sync_ref(self, mirror: Dict, ref: Reference, skip_ci: bool, dry_run: bool) -> None:  # noqa: WPS210,WPS231
        """Synchronize the given branch or tag.

        Args:
            mirror: Current mirror dict.
            ref: reference objects.
            skip_ci: skip_ci push option.
            dry_run: Dry run.
        """
        if ref.name.startswith('refs/heads/'):
            ref_type = 'head'
            ref_type_human = 'branch'
            ref_type_human_plural = 'branches'
        elif ref.name.startswith('refs/tags/'):
            ref_type = 'tag'
            ref_type_human = 'tag'
            ref_type_human_plural = 'tags'
        else:
            return
        shorthand = ref.name.split('/', 2)[2]

        # Ref mapping
        dest_shortand = shorthand
        if ref_type_human_plural in mirror:
            dest_shortand = None
            mappings: List[Dict[str, str]] = mirror.get(ref_type_human_plural)  # type: ignore
            for mapping in mappings:
                matcher = Matcher(
                    mapping.get('from', ''),
                    log_prefix='[{0}] {1} {2}'.format(
                        self._name,  # type: ignore
                        mirror['url'],
                        ref_type_human_plural,
                    ),
                )
                matches = matcher.match([shorthand])
                if matches:
                    to_param = mapping.get('to', shorthand)
                    dest_shortand = matches[0].expand(to_param)
                    break

        if dest_shortand is None:
            return

        if skip_ci:
            # Note: Ignored by libgit2/PyGit2
            # https://github.com/libgit2/libgit2/issues/5335
            self._repo.config['push.pushOption'] = 'ci.skip'
        else:
            self._repo.config['push.pushOption'] = ''

        pull_commit = ref.peel().id
        gitlab_ref = self._repo.references.get(
            'refs/remotes/gitlab/{ref_type}s/{ref}'.format(ref_type=ref_type, ref=dest_shortand),
        )
        try:
            gitlab_commit = gitlab_ref.peel().id
        except AttributeError:
            gitlab_commit = None
        if pull_commit != gitlab_commit:
            if dry_run:
                logger.info(
                    '[%s] %s NOT Pushing %s %s to %s: %s -> %s (dry-run)',
                    self._name,  # type: ignore
                    mirror['url'],
                    ref_type_human,
                    shorthand,
                    dest_shortand,
                    gitlab_commit,
                    str(pull_commit),
                )
            else:
                logger.info(
                    '[%s] %s Pushing %s %s to %s: %s -> %s',
                    self._name,  # type: ignore
                    mirror['url'],
                    ref_type_human,
                    shorthand,
                    dest_shortand,
                    gitlab_commit,
                    str(pull_commit),
                )
                refspec = '{ref_name}:refs/{ref_type}s/{ref}'.format(
                    ref_name=ref.name,
                    ref_type=ref_type,
                    ref=dest_shortand,
                )
                self._push_remote(
                    'gitlab',
                    [refspec],
                    self.pygit2_credentials,  # type: ignore
                )

    def _pull_mirror(self, mirror: Dict, skip_ci: bool, dry_run: bool) -> None:
        """Pull from the given mirror and push.

        Args:
            mirror: Current mirror dict.
            skip_ci: skip_ci push option.
            dry_run: Dry run.
        """
        try:
            self._repo.remotes['pull']  # noqa: WPS428
        except KeyError:
            self._repo.remotes.create('pull', mirror['url'], '+refs/heads/*:refs/heads/*')
            self._repo.remotes.add_fetch('pull', '+refs/tags/*:refs/tags/*')
            self._repo.config['remote.pull.mirror'] = True
        credentials = None
        pull_auth_id = mirror.get('auth_id')
        if pull_auth_id:
            credentials = GitlabConnections().get_connection(pull_auth_id).pygit2_credentials
        self._fetch_remote('pull', credentials)
        for ref in self._repo.references.objects:
            self._sync_ref(mirror, ref, skip_ci, dry_run)

    def _process_mirrors(
        self,
        param_name: str,
        param_value: Any,
        dry_run: bool = False,
        skip_save: bool = False,
    ) -> None:
        """Process the mirrors param.

        Args:
            param_name: "mirrors".
            param_value: List of mirror dicts.
            dry_run: Dry run.
            skip_save: False.
        """
        assert param_name == 'mirrors'  # noqa: S101
        assert not skip_save  # noqa: S101

        pull_mirror_count = 0
        self._init_repo()
        self._fetch_remote(
            'gitlab',
            self.pygit2_credentials,  # type: ignore
        )
        for mirror in param_value:
            direction = mirror.get('direction', 'pull')
            skip_ci = mirror.get('skip_ci', True)
            if direction == 'pull':
                if pull_mirror_count > 0:
                    logger.warning(
                        '[%s] NOT Pulling mirror: %s (Only first pull mirror is processed)',
                        self._name,  # type: ignore
                        mirror['url'],
                    )
                    continue
                self._pull_mirror(mirror, skip_ci, dry_run)
                pull_mirror_count += 1
            else:
                logger.warning(
                    '[%s] NOT Push mirror: %s (Not supported yet)',
                    self._name,  # type: ignore
                    mirror['url'],
                )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import logging
from pathlib import PurePosixPath
from typing import TYPE_CHECKING, Match, Optional, Tuple

from gitlabracadabra.containers.registries import ReferenceParts, Registries
from gitlabracadabra.matchers import Matcher


try:
    from semantic_version import NpmSpec, Version  # noqa: WPS433
except ImportError:
    # semantic_version < 2.7
    from semantic_version import Spec as NpmSpec, Version  # noqa: WPS433,WPS440

if TYPE_CHECKING:
    from typing import Any, Dict, List
    from gitlabracadabra.containers.registry import Registry


logger = logging.getLogger(__name__)

Source = Tuple[PurePosixPath, str, Match, Optional[str]]


class ImageMirrorsMixin(object):  # noqa: WPS214
    """Object (Project) with image mirrors."""

    def _process_image_mirrors(
        self,
        param_name: str,
        param_value: Any,
        dry_run: bool = False,
        skip_save: bool = False,
    ) -> None:
        """Process the image_mirrors param.

        Args:
            param_name: "image_mirrors".
            param_value: List of image mirror dicts.
            dry_run: Dry run.
            skip_save: False.
        """
        assert param_name == 'image_mirrors'  # noqa: S101
        assert not skip_save  # noqa: S101

        dest_registry, prefix = self._get_destination()

        for image_mirror in param_value:
            self._mirror(image_mirror, dest_registry, prefix, dry_run=dry_run)

    def _get_destination(self) -> Tuple[Registry, PurePosixPath]:
        # https://gitlab.com/gitlab-org/gitlab/-/merge_requests/54090
        try:
            container_registry_image_prefix = self._obj.container_registry_image_prefix  # type: ignore
        except AttributeError:
            container_registry_image_prefix = None
        if not isinstance(container_registry_image_prefix, str):
            raise ValueError('Unexpected type for container_registry_image_prefix: {0}'.format(
                type(container_registry_image_prefix),
            ))

        netloc, path = container_registry_image_prefix.split('/', 1)
        registry = Registries().get_registry(netloc, self.registry_auth_info)  # type: ignore
        return (registry, PurePosixPath(path))

    def _mirror(  # noqa: WPS210
        self,
        image_mirror: Dict,
        dest_registry: Registry,
        prefix: PurePosixPath,
        *,
        dry_run: bool,
    ) -> None:
        semver = None
        if 'semver' in image_mirror:
            semver = NpmSpec(image_mirror.get('semver'))
        sources = self._get_sources(image_mirror.get('from'), semver)
        for source in sources:
            if not self._match_semver(source, semver):
                continue
            dest = self._get_dest(
                source[0],
                source[1],
                source[2],
                source[3],
                image_mirror.get('to'),
            )
            digest_suffix = ''
            if source[3]:
                digest_suffix = '@{0}'.format(source[3])
            source_manifest = Registries().get_manifest('{0}:{1}{2}'.format(
                str(source[0] / source[1]),
                source[2][0],
                digest_suffix,
            ))
            dest_manifest = Registries().get_manifest(ReferenceParts(
                dest_registry.hostname,
                str(prefix / dest[0]),
                dest[1],
                None,
            ))
            dest_manifest.registry.import_manifest(
                source_manifest,
                dest_manifest.manifest_name,
                tag=dest_manifest.tag,
                platform=None,
                log_prefix='[{0}] '.format(self._name),  # type: ignore
                dry_run=dry_run,
            )

    def _get_sources(self, from_param: Any, semver: Optional[NpmSpec]) -> List[Source]:
        """Get sources.

        Args:
            from_param: The "from" param.
            semver: Optional NpmSpec.

        Returns:
            A list of tuples (base, repository, tag, digest)

        Raises:
            ValueError: Unexpected from param type.
        """
        default_tag = 'latest'
        if semver:
            default_tag = '/.*/'
        if isinstance(from_param, str):
            parts = Registries().full_reference_parts(from_param)
            return self._get_sources_from_parts(
                PurePosixPath(parts.hostname),
                [parts.manifest_name],
                [parts.tag or default_tag],
                parts.digest,
            )
        elif isinstance(from_param, dict):
            base = PurePosixPath(from_param.get('base', ''))
            repositories = from_param.get('repositories', None)
            tags = from_param.get('tags', [default_tag])
            return self._get_sources_from_parts(base, repositories, tags, None)
        raise ValueError('Unexpected from param type: {0}'.format(type(from_param)))

    def _get_sources_from_parts(
        self,
        base: PurePosixPath,
        repositories: List[str],
        tags: List[str],
        digest: Optional[str],
    ) -> List[Source]:
        if not isinstance(repositories, list):
            raise ValueError('Unexpected from.repositories param type: {0}'.format(type(repositories)))
        if not isinstance(tags, list):
            raise ValueError('Unexpected from.tags param type: {0}'.format(type(tags)))
        sources: List[Source] = []
        for repository in repositories:
            matcher = Matcher(
                tags,
                log_prefix='[{0}] '.format(self._name),  # type: ignore
            )
            manifest = Registries().get_manifest(str(base / repository))
            for match in matcher.match(manifest.tag_list):
                sources.append((base, repository, match, digest))
        return sources

    def _match_semver(
        self,
        source: Source,
        semver: Optional[NpmSpec],
    ) -> bool:
        if not semver:
            return True
        version_str = source[2].group(0)
        if version_str.startswith('v'):
            version_str = version_str[1:]
        try:
            version = Version(version_str)
        except ValueError as err:
            logger.warning(  # noqa: G200
                '[%s] %s for %s',
                self._name,  # type: ignore
                str(err),
                str(source[0] / source[1]),
            )
            return False
        return version in semver

    def _get_dest(  # noqa: WPS211
        self,
        source_base: PurePosixPath,
        source_repository: str,
        source_tag: Match,
        source_digest: Optional[str],
        to_param: Any,
    ) -> Tuple[str, str]:
        """Get Destination.

        Args:
            source_base: base as PurePosixPath.
            source_repository: source repository, relative to base.
            source_tag: source tag.
            source_digest: source digest.
            to_param: "to" parameter.

        Returns:
            A tuple (repository, tag)

        Raises:
            ValueError: Unexpected to param type.
        """
        if to_param is None:
            repository = str(source_base / source_repository).split('/', 1).pop()
            return repository, source_tag[0]
        elif isinstance(to_param, str):
            tag: Optional[str]
            try:
                to_param, tag = to_param.rsplit(':', 1)
            except ValueError:
                tag = None
            return to_param, tag or source_tag[0]
        elif isinstance(to_param, dict):
            return self._get_dest_from_dict(source_base, source_repository, source_tag, source_digest, to_param)
        raise ValueError('Unexpected to param type: {0}'.format(type(to_param)))

    def _get_dest_from_dict(  # noqa: WPS211
        self,
        source_base: PurePosixPath,
        source_repository: str,
        source_tag: Match,
        source_digest: Optional[str],
        to_param: Dict,
    ) -> Tuple[str, str]:
        base = PurePosixPath(to_param.get('base', source_base) or '')
        repository = to_param.get('repository', source_repository) or ''
        tag = source_tag.expand(to_param.get('tag') or source_tag[0])
        return str(base / repository), tag



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import namedtuple
from unittest.mock import MagicMock, call, patch

from semantic_version import __version__ as semantic_version_version

from gitlabracadabra.containers.registries import Registries
from gitlabracadabra.containers.registry import Registry
from gitlabracadabra.containers.with_digest import WithDigest
from gitlabracadabra.objects.project import GitLabracadabraProject
from gitlabracadabra.tests import my_vcr
from gitlabracadabra.tests.case import TestCaseWithManager


VERIFY_DIGEST_METHOD = '_verify_digest'

TestData = namedtuple('TestData', ['mirrors', 'mappings'])

TEST_DATA = (
    TestData(
        [{'from': 'debian'}],
        (('docker.io/library/debian:latest', 'mygroup/myproject/library/debian:latest'),),
    ),
    TestData(
        [{'from': 'debian:buster'}],
        (('docker.io/library/debian:buster', 'mygroup/myproject/library/debian:buster'),),
    ),
    TestData(
        [{'from': 'debian:buster', 'to': ''}],
        (('docker.io/library/debian:buster', 'mygroup/myproject:buster'),),
    ),
    TestData(
        [{'from': 'debian:buster', 'to': ':latest'}],
        (('docker.io/library/debian:buster', 'mygroup/myproject:latest'),),
    ),
    TestData(
        [{'from': 'quay.org/coreos/etcd:v3.4.1', 'to': 'etcd:v3.4.1'}],
        (('quay.org/coreos/etcd:v3.4.1', 'mygroup/myproject/etcd:v3.4.1'),),
    ),
    TestData(
        [{'from': {'repositories': ['a']}, 'to': 'etcd:v3.4.1'}],
        (('docker.io/library/a:latest', 'mygroup/myproject/etcd:v3.4.1'),),
    ),
    TestData(
        [{'from': {'repositories': ['a', 'b']}, 'to': 'etcd:v3.4.1'}],
        (
            ('docker.io/library/a:latest', 'mygroup/myproject/etcd:v3.4.1'),
            ('docker.io/library/b:latest', 'mygroup/myproject/etcd:v3.4.1'),
        ),
    ),
    TestData(
        [{'from': {
            'base': 'k8s.gcr.io/kubernetes',
            'repositories': ['kube-apiserver', 'kube-proxy'],
            'tags': ['v1.20.4', 'v1.20.6'],
        }}],
        (
            ('k8s.gcr.io/kubernetes/kube-apiserver:v1.20.4', 'mygroup/myproject/kubernetes/kube-apiserver:v1.20.4'),
            ('k8s.gcr.io/kubernetes/kube-apiserver:v1.20.6', 'mygroup/myproject/kubernetes/kube-apiserver:v1.20.6'),
            ('k8s.gcr.io/kubernetes/kube-proxy:v1.20.4', 'mygroup/myproject/kubernetes/kube-proxy:v1.20.4'),
            ('k8s.gcr.io/kubernetes/kube-proxy:v1.20.6', 'mygroup/myproject/kubernetes/kube-proxy:v1.20.6'),
        ),
    ),
    TestData(
        [{
            'from': {
                'base': 'k8s.gcr.io/kubernetes',
                'repositories': ['kube-apiserver', 'kube-proxy'],
                'tags': ['v1.20.4', 'v1.20.6'],
            },
            'to': {},
        }],
        (
            ('k8s.gcr.io/kubernetes/kube-apiserver:v1.20.4', 'mygroup/myproject/k8s.gcr.io/kubernetes/kube-apiserver:v1.20.4'),
            ('k8s.gcr.io/kubernetes/kube-apiserver:v1.20.6', 'mygroup/myproject/k8s.gcr.io/kubernetes/kube-apiserver:v1.20.6'),
            ('k8s.gcr.io/kubernetes/kube-proxy:v1.20.4', 'mygroup/myproject/k8s.gcr.io/kubernetes/kube-proxy:v1.20.4'),
            ('k8s.gcr.io/kubernetes/kube-proxy:v1.20.6', 'mygroup/myproject/k8s.gcr.io/kubernetes/kube-proxy:v1.20.6'),
        ),
    ),
    TestData(
        [{
            'from': {
                'base': 'k8s.gcr.io/kubernetes',
                'repositories': ['kube-apiserver', 'kube-proxy'],
                'tags': ['v1.20.4', 'v1.20.6'],
            },
            'to': {
                'base': 'new_base',
            },
        }],
        (
            ('k8s.gcr.io/kubernetes/kube-apiserver:v1.20.4', 'mygroup/myproject/new_base/kube-apiserver:v1.20.4'),
            ('k8s.gcr.io/kubernetes/kube-apiserver:v1.20.6', 'mygroup/myproject/new_base/kube-apiserver:v1.20.6'),
            ('k8s.gcr.io/kubernetes/kube-proxy:v1.20.4', 'mygroup/myproject/new_base/kube-proxy:v1.20.4'),
            ('k8s.gcr.io/kubernetes/kube-proxy:v1.20.6', 'mygroup/myproject/new_base/kube-proxy:v1.20.6'),
        ),
    ),
    TestData(
        [{
            'from': {
                'base': 'k8s.gcr.io/kubernetes',
                'repositories': ['kube-apiserver', 'kube-proxy'],
                'tags': ['v1.20.4', 'v1.20.6'],
            },
            'to': {
                'repository': 'new_repo',
            },
        }],
        (
            ('k8s.gcr.io/kubernetes/kube-apiserver:v1.20.4', 'mygroup/myproject/k8s.gcr.io/kubernetes/new_repo:v1.20.4'),
            ('k8s.gcr.io/kubernetes/kube-apiserver:v1.20.6', 'mygroup/myproject/k8s.gcr.io/kubernetes/new_repo:v1.20.6'),
            ('k8s.gcr.io/kubernetes/kube-proxy:v1.20.4', 'mygroup/myproject/k8s.gcr.io/kubernetes/new_repo:v1.20.4'),
            ('k8s.gcr.io/kubernetes/kube-proxy:v1.20.6', 'mygroup/myproject/k8s.gcr.io/kubernetes/new_repo:v1.20.6'),
        ),
    ),
    TestData(
        [{
            'from': {
                'base': 'k8s.gcr.io/kubernetes',
                'repositories': ['kube-apiserver', 'kube-proxy'],
                'tags': ['v1.20.4', 'v1.20.6'],
            },
            'to': {
                'tag': 'new_tag',
            },
        }],
        (
            ('k8s.gcr.io/kubernetes/kube-apiserver:v1.20.4', 'mygroup/myproject/k8s.gcr.io/kubernetes/kube-apiserver:new_tag'),
            ('k8s.gcr.io/kubernetes/kube-apiserver:v1.20.6', 'mygroup/myproject/k8s.gcr.io/kubernetes/kube-apiserver:new_tag'),
            ('k8s.gcr.io/kubernetes/kube-proxy:v1.20.4', 'mygroup/myproject/k8s.gcr.io/kubernetes/kube-proxy:new_tag'),
            ('k8s.gcr.io/kubernetes/kube-proxy:v1.20.6', 'mygroup/myproject/k8s.gcr.io/kubernetes/kube-proxy:new_tag'),
        ),
    ),
    # Regexps
    TestData(
        [{
            'from': {
                'repositories': ['busybox', 'debian'],
                'tags': ['/(uns|s)(id|table)/'],
            },
            'to': {
                'tag': r'how-\1-\2',
            },
        }],
        (
            ('docker.io/library/busybox:stable', 'mygroup/myproject/busybox:how-s-table'),
            ('docker.io/library/busybox:unstable', 'mygroup/myproject/busybox:how-uns-table'),
            ('docker.io/library/debian:sid', 'mygroup/myproject/debian:how-s-id'),
            ('docker.io/library/debian:stable', 'mygroup/myproject/debian:how-s-table'),
            ('docker.io/library/debian:unstable', 'mygroup/myproject/debian:how-uns-table'),
        ),
    ),
    TestData(
        [{
            'from': 'busybox:/(uns|s)(id|table)/',
        }],
        (
            ('docker.io/library/busybox:stable', 'mygroup/myproject/library/busybox:stable'),
            ('docker.io/library/busybox:unstable', 'mygroup/myproject/library/busybox:unstable'),
        ),
    ),
    TestData(
        [{
            'from': 'busybox:/(uns|s)(id|table)/',
            'to': {
                'tag': r'how-\1-\2',
            },
        }],
        (
            ('docker.io/library/busybox:stable', 'mygroup/myproject/docker.io/library/busybox:how-s-table'),
            ('docker.io/library/busybox:unstable', 'mygroup/myproject/docker.io/library/busybox:how-uns-table'),
        ),
    ),
    # Semver
    TestData(
        [{
            'from': 'k8s.gcr.io/kubernetes/kube-apiserver',
            'semver': '>=1.20.5',
        }],
        (
            ('k8s.gcr.io/kubernetes/kube-apiserver:v1.20.5', 'mygroup/myproject/kubernetes/kube-apiserver:v1.20.5'),
            ('k8s.gcr.io/kubernetes/kube-apiserver:v1.21.0', 'mygroup/myproject/kubernetes/kube-apiserver:v1.21.0'),
        ),
    ),
)


class TestProjectImageMirrors(TestCaseWithManager):
    """Test image_mirrors param."""

    @my_vcr.use_cassette
    def test_image_mirrors_mocked(self, cass):
        """Test image_mirrors, with mocked Registry.import_manifest.

        Args:
            cass: VCR cassette.
        """
        registry = Registries().get_registry('registry.example.com')
        for test_data in TEST_DATA:
            with self.subTest(mirrors=test_data.mirrors):
                with patch.object(Registry, 'import_manifest', autospec=True) as import_manifest_mock:
                    project = GitLabracadabraProject('memory', 'mygroup/myproject', {
                        'image_mirrors': test_data.mirrors,
                    })
                    self.assertEqual(project.errors(), [])
                    project._get = MagicMock()  # noqa: WPS437
                    project._obj = type('', (), {})()  # noqa: WPS437
                    project._obj.container_registry_image_prefix = 'registry.example.com/mygroup/myproject'  # noqa: WPS437
                    project.process()
                    if any('semver' in mirror for mirror in test_data.mirrors) and semantic_version_version == '2.6.0':
                        self.skipTest('Too old semantic_version version={0}'.format(semantic_version_version))  # noqa: WPS220
                        continue  # noqa: WPS220
                    self._assert_mappings(registry, import_manifest_mock, test_data.mappings)
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_image_mirrors_auth(self, cass):
        """Test image_mirrors, auth'ing using JWT.

        Args:
            cass: VCR cassette.
        """
        expected_digest = 'sha256:c6b45a95f932202dbb27c31333c4789f45184a744060f6e569cc9d2bf1b9ad6f'
        with patch.object(WithDigest, VERIFY_DIGEST_METHOD) as verify_digest_mock:
            verify_digest_mock.return_value = None
            project = GitLabracadabraProject('memory', 'root/test_registry', {
                'image_mirrors': [{
                    'from': 'busybox',
                    'to': 'busybox',
                }],
            })
            self.assertEqual(project.errors(), [])
            with patch('gitlabracadabra.containers.registry_importer.logger', autospec=True) as logger:
                project.process()
                logger.info.assert_has_calls([call(
                    '%s%s %simported as %s:%s (%s, %s, %s)',
                    '[root/test_registry] ',
                    'docker.io/library/busybox:latest@{0}'.format(expected_digest),
                    '',
                    'root/test_registry/busybox',
                    'latest',
                    '1+0=1 uploaded+existing manifests',
                    '2+0+0=2 uploaded+mounted+existing blobs',
                    '766112+0+0=766112 uploaded+mounted+existing blobs size',
                )])
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_image_mirror_digest(self, cass):
        """Test image_mirrors, with source digest.

        Args:
            cass: VCR cassette.
        """
        expected_digest = 'sha256:de396b540b82219812061d0d753440d5655250c621c753ed1dc67d6154741607'
        with patch.object(WithDigest, VERIFY_DIGEST_METHOD) as verify_digest_mock:
            verify_digest_mock.return_value = None
            project = GitLabracadabraProject('memory', 'root/test_registry', {
                'image_mirrors': [{
                    'from': 'quay.io/operator-framework/olm@{0}'.format(expected_digest),
                }],
            })
            self.assertEqual(project.errors(), [])
            with patch('gitlabracadabra.containers.registry_importer.logger', autospec=True) as logger:
                project.process()
                logger.info.assert_has_calls([call(
                    '%s%s %simported as %s:%s (%s, %s, %s)',
                    '[root/test_registry] ',
                    'quay.io/operator-framework/olm:latest@{0}'.format(expected_digest),
                    '',
                    'root/test_registry/operator-framework/olm',
                    'latest',
                    '1+0=1 uploaded+existing manifests',
                    '1+0+5=6 uploaded+mounted+existing blobs',
                    '16792050+0+76720801=93512851 uploaded+mounted+existing blobs size',
                )])
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_image_mirror_no_digest_header(self, cass):
        """Test image_mirrors, without Docker-Content-Digest header.

        Args:
            cass: VCR cassette.
        """
        expected_digest = 'sha256:7b4f6f33cdbdeaa631fc0ef53050d6880749acd965cd655e4a6b7604a027ff91'
        with patch.object(WithDigest, VERIFY_DIGEST_METHOD) as verify_digest_mock:
            verify_digest_mock.return_value = None
            project = GitLabracadabraProject('memory', 'root/test_registry', {
                'image_mirrors': [{
                    'from': 'registry.access.redhat.com/rhscl/postgresql-10-rhel7:1',
                }],
            })
            self.assertEqual(project.errors(), [])
            with patch('gitlabracadabra.containers.registry_importer.logger', autospec=True) as logger:
                project.process()
                logger.info.assert_has_calls([call(
                    '%s%s %simported as %s:%s (%s, %s, %s)',
                    '[root/test_registry] ',
                    'registry.access.redhat.com/rhscl/postgresql-10-rhel7:1@{0}'.format(expected_digest),
                    '',
                    'root/test_registry/rhscl/postgresql-10-rhel7',
                    '1',
                    '1+0=1 uploaded+existing manifests',
                    '4+0+1=5 uploaded+mounted+existing blobs',
                    '117503892+0+6812=117510704 uploaded+mounted+existing blobs size',
                )])
        self.assertTrue(cass.all_played)

    @my_vcr.use_cassette
    def test_image_mirror_manifest_v1(self, cass):
        """Test image_mirrors, application/vnd.docker.distribution.manifest.v1+prettyjws.

        Args:
            cass: VCR cassette.
        """
        expected_digest = 'sha256:b19e1c859c5d1b23169453a5abfaa37299106f77a1e1227381739c4461915984'
        with patch.object(WithDigest, VERIFY_DIGEST_METHOD) as verify_digest_mock:
            verify_digest_mock.return_value = None
            project = GitLabracadabraProject('memory', 'root/test_registry', {
                'image_mirrors': [{
                    'from': 'quay.io/jetstack/cert-manager-controller:v0.1.0',
                }],
            })
            self.assertEqual(project.errors(), [])
            with patch('gitlabracadabra.containers.registry_importer.logger', autospec=True) as logger:
                project.process()
                logger.info.assert_has_calls([call(
                    '%s%s %simported as %s:%s (%s, %s, %s)',
                    '[root/test_registry] ',
                    'quay.io/jetstack/cert-manager-controller:v0.1.0@{0}'.format(expected_digest),
                    '',
                    'root/test_registry/jetstack/cert-manager-controller',
                    'v0.1.0',
                    '1+0=1 uploaded+existing manifests',
                    '4+0+3=7 uploaded+mounted+existing blobs',
                    '56+0+42=98 uploaded+mounted+existing blobs size',  # 14173208+0+96=14173304
                )])

    @my_vcr.use_cassette
    def test_image_mirrors_mounted(self, cass):
        """Test image_mirrors, with mounted blobs.

        Args:
            cass: VCR cassette.
        """
        expected_digest = 'sha256:37aafd8a0b88f753556b2dc89091671f8bedeb869b2acf9aa2a22bffd563807a'
        with patch.object(WithDigest, VERIFY_DIGEST_METHOD) as verify_digest_mock:
            verify_digest_mock.return_value = None
            project = GitLabracadabraProject('memory', 'root/test_registry', {
                'image_mirrors': [
                    {'from': 'registry.developers.crunchydata.com/crunchydata/pgo-apiserver:centos8-4.6.2'},
                    {'from': 'registry.developers.crunchydata.com/crunchydata/pgo-event:centos8-4.6.2'},
                ],
            })
            self.assertEqual(project.errors(), [])
            with patch('gitlabracadabra.containers.registry_importer.logger', autospec=True) as logger:
                project.process()
                logger.info.assert_has_calls([call(
                    '%s%s %simported as %s:%s (%s, %s, %s)',
                    '[root/test_registry] ',
                    'registry.developers.crunchydata.com/crunchydata/pgo-event:centos8-4.6.2@{0}'.format(expected_digest),
                    '',
                    'root/test_registry/crunchydata/pgo-event',
                    'centos8-4.6.2',
                    '1+0=1 uploaded+existing manifests',
                    '2+11+2=15 uploaded+mounted+existing blobs',
                    '18993004+121840748+64=140833816 uploaded+mounted+existing blobs size',
                )])
        self.assertTrue(cass.all_played)

    def _assert_mappings(self, registry, import_manifest_mock, mappings):
        self.assertEqual(len(import_manifest_mock.mock_calls), len(mappings))
        for mapping_index, mapping in enumerate(mappings):
            _, args, kwargs = import_manifest_mock.mock_calls[mapping_index]
            self.assertEqual(len(args), 3)
            self.assertEqual(args[0], registry)
            self.assertEqual(str(args[1]), mapping[0])
            self.assertEqual(
                '{0}:{1}'.format(args[2], kwargs['tag']),
                mapping[1],
            )
            self.assertEqual(
                kwargs,
                {
                    'tag': mapping[1].rsplit(':').pop(),
                    'platform': None,
                    'log_prefix': '[mygroup/myproject] ',
                    'dry_run': False,
                },
            )
